__all__ = ()

from scarletio import include, include_with_callback

from ...application import Application
from ...bases import id_sort_key
from ...embed import Embed
from ...emoji import ReactionMapping, merge_update_reaction_mapping
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory, flag_parser_factory,
    nullable_date_time_parser_factory, nullable_entity_array_parser_factory, nullable_entity_parser_factory,
    nullable_functional_array_parser_factory, nullable_object_array_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    flag_optional_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_entity_array_optional_putter_factory, nullable_entity_optional_putter_factory,
    nullable_functional_array_optional_putter_factory, nullable_object_array_optional_putter_factory,
    nullable_string_optional_putter_factory, optional_entity_id_array_optional_putter_factory,
    preinstanced_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator_factory, entity_id_array_validator_factory,
    entity_id_validator_factory, flag_validator_factory, nullable_date_time_validator_factory,
    nullable_entity_array_validator_factory, nullable_entity_validator_factory, nullable_object_array_validator_factory,
    nullable_string_validator_factory, preinstanced_validator_factory
)
from ...poll import Poll
from ...role import Role
from ...soundboard import SoundboardSound
from ...sticker import Sticker, create_partial_sticker_data, create_partial_sticker_from_partial_data
from ...user import ClientUserBase, User, UserBase, ZEROUSER
from ...webhook import WebhookBase, WebhookRepr, WebhookType, create_partial_webhook_from_id

from ..attachment import Attachment
from ..message_activity import MessageActivity
from ..message_application import MessageApplication
from ..message_call import MessageCall
from ..message_interaction import MessageInteraction
from ..message_role_subscription import MessageRoleSubscription
from ..poll_change import PollChange
from ..poll_update import PollUpdate
from ..shared_client_theme import SharedClientTheme

from .constants import CONTENT_LENGTH_MAX, NONCE_LENGTH_MAX
from .flags import MessageFlag
from .preinstanced import MessageType


Channel = include('Channel')
Component = include('Component')
Message = include('Message')
MessageSnapshot = include('MessageSnapshot')
Resolved = include('Resolved')


# activity

parse_activity = nullable_entity_parser_factory('activity', MessageActivity)
put_activity = nullable_entity_optional_putter_factory('activity', MessageActivity)
validate_activity = nullable_entity_validator_factory('activity', MessageActivity)

# application

parse_application = nullable_entity_parser_factory('application', MessageApplication)
put_application = nullable_entity_optional_putter_factory(
    'application', MessageApplication, force_include_internals = True
)
validate_application = nullable_entity_validator_factory('application', MessageApplication)

# application_id

parse_application_id = entity_id_parser_factory('application_id')


def put_application_id(application_id, data, defaults):
    """
    Puts the message's application's identifier into the given `data` json serializable object.
    
    Parameters
    ----------
    application_id : `int`
        Application identifier.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if application_id:
        raw_application_id = str(application_id)
        data['application_id'] = raw_application_id
        
        if data.get('webhook_id', None) is None:
            data['webhook_id'] = raw_application_id
    
    elif defaults:
        data['application_id'] = None
    
    return data

validate_application_id = entity_id_validator_factory('application_id', Application)

# attachments

parse_attachments = nullable_entity_array_parser_factory('attachments', Attachment)
put_attachments = nullable_entity_array_optional_putter_factory('attachments', Attachment)
validate_attachments = nullable_entity_array_validator_factory('attachments', Attachment)

# author

def parse_author(data, guild_id = 0, channel_id = 0):
    """
    Parses the message's author from its data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's id where the message was created at.
    channel_id : `int` = `0`, Optional
        The channel's id where the message was created at.
    
    Returns
    -------
    author : ``UserBase``
    """
    author_data = data.get('author', None)
    
    webhook_id = data.get('webhook_id', None)
    if (webhook_id is not None):
        application_id = data.get('application_id', None)
        if ((application_id is None) or (webhook_id != application_id)):
            if data.get('message_reference', None) is None:
                webhook_type = WebhookType.bot
            else:
                webhook_type = WebhookType.server
            
            webhook_id = int(webhook_id)
            if author_data is None:
                return create_partial_webhook_from_id(webhook_id, '', webhook_type = webhook_type)
            
            return WebhookRepr.from_data(author_data, webhook_id, webhook_type, channel_id)
    
    if author_data is None:
        return ZEROUSER
    
    return User.from_data(author_data, data.get('member', None), guild_id)


def put_author(author, data, defaults, *, guild_id = 0):
    """
    Puts the message author's data into the given `data` json serializable object.
    
    Parameters
    ----------
    author : ``UserBase``
        Message author.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['author'] =  author.to_data(defaults = defaults, include_internals = True)
    
    if guild_id:
        try:
            guild_profile = author.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
    
    # If the message as created by a webhook, put its into the data too.
    if isinstance(author, WebhookBase):
        data['webhook_id'] = str(author.id)
    elif defaults:
        data['webhook_id'] = None
    
    return data


validate_author = default_entity_validator_factory('author', UserBase, default = ZEROUSER)

# call

parse_call = nullable_entity_parser_factory('call', MessageCall)
put_call = nullable_entity_optional_putter_factory('call', MessageCall)
validate_call = nullable_entity_validator_factory('call', MessageCall)

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', NotImplemented, include = 'Channel')

# components

COMPONENT_TYPE_LAYOUT_MAS_REQUIRED = 0

@include_with_callback('ComponentTypeLayoutFlag')
def _set_COMPONENT_TYPE_LAYOUT_MAS_REQUIRED(ComponentTypeLayoutFlag):
    global COMPONENT_TYPE_LAYOUT_MAS_REQUIRED
    COMPONENT_TYPE_LAYOUT_MAS_REQUIRED = ComponentTypeLayoutFlag().update_by_keys(
        allowed_in_message = True,
        top_level = True,
    )

del _set_COMPONENT_TYPE_LAYOUT_MAS_REQUIRED


parse_components = nullable_object_array_parser_factory('components', NotImplemented, include = 'Component')
put_components = nullable_object_array_optional_putter_factory('components', can_include_internals = True)


def validate_components(components):
    """
    Validates the given components.
    
    Parameters
    ----------
    components : ``None | iterable<Component | (tuple | list)<Component>>``
        The components to validate.
    
    Returns
    -------
    components_processed : ``None | tuple<Component>``
    
    Raises
    ------
    TypeError
        - If `components` is invalid type.
    ValueError
        - If `components` contains an invalid value.
    """
    if components is None:
        return None
    
    if (getattr(components, '__iter__', None) is None):
        raise TypeError(
            f'`components` can be `None`, `iterable` of `{Component.__name__}`, got '
            f'{type(components).__name__}; {components!r}.'
        )
        
    components_processed = None
    
    for component in components:
        if not isinstance(component, Component):
            raise TypeError(
                f'`components` can contain `{Component.__name__}` elements, got '
                f'{type(component).__name__}; {component!r}; components = {components!r}.'
            )
        
        if component.type.layout_flags & COMPONENT_TYPE_LAYOUT_MAS_REQUIRED != COMPONENT_TYPE_LAYOUT_MAS_REQUIRED:
            raise ValueError(
                f'Component of type {component.type.name}, cannot be used as a top level component in messages, '
                f'got {component!r}; components = {components!r}.'
            )
        
        if (components_processed is None):
            components_processed = []
        
        components_processed.append(component)
    
    if (components_processed is not None):
        components_processed = tuple(components_processed)
    
    return components_processed


# content

parse_content = nullable_string_parser_factory('content')
put_content = nullable_string_optional_putter_factory('content')
validate_content = nullable_string_validator_factory('content', 0, CONTENT_LENGTH_MAX)

# embed

parse_embeds = nullable_object_array_parser_factory('embeds', Embed)
put_embeds = nullable_object_array_optional_putter_factory('embeds', Embed)
validate_embeds = nullable_object_array_validator_factory('embeds', Embed)

# edited_at

parse_edited_at = nullable_date_time_parser_factory('edited_timestamp')
put_edited_at = nullable_date_time_optional_putter_factory('edited_timestamp')
validate_edited_at = nullable_date_time_validator_factory('edited_at')

# flags

parse_flags = flag_parser_factory('flags', MessageFlag)
put_flags = flag_optional_putter_factory('flags', MessageFlag())
validate_flags = flag_validator_factory('flags', MessageFlag)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('message_id')

# interaction

# Old messages use `interaction` instead of `interaction_metadata`.
def parse_interaction(data):
    """
    Parses message interaction from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    interaction : `None | MessageInteraction`
    """
    try:
        interaction_data = data['interaction_metadata']
    except KeyError:
        interaction_data = data.get('interaction', None)
    
    if (interaction_data is not None):
        return MessageInteraction.from_data(interaction_data)


put_interaction = nullable_entity_optional_putter_factory(
    'interaction_metadata', MessageInteraction, force_include_internals = True
)
validate_interaction = nullable_entity_validator_factory('interaction', MessageInteraction)

# mentioned_channels_cross_guild

parse_mentioned_channels_cross_guild = nullable_functional_array_parser_factory(
    'mention_channels',
    NotImplemented,
    include = 'create_partial_channel_from_data',
    sort_key = id_sort_key,
)
put_mentioned_channels_cross_guild = nullable_functional_array_optional_putter_factory(
    'mention_channels',
    NotImplemented,
    include = 'create_partial_channel_data',
)
validate_mentioned_channels_cross_guild = nullable_entity_array_validator_factory(
    'mentioned_channels_cross_guild', NotImplemented, include = 'Channel', sort_key = id_sort_key,
)

# mentioned_everyone

parse_mentioned_everyone = bool_parser_factory('mention_everyone', False)
put_mentioned_everyone = bool_optional_putter_factory('mention_everyone', False)
validate_mentioned_everyone = bool_validator_factory('mention_everyone', False)

# mentioned_role_ids

parse_mentioned_role_ids = entity_id_array_parser_factory('mention_roles')
put_mentioned_role_ids = optional_entity_id_array_optional_putter_factory('mention_roles')
validate_mentioned_role_ids = entity_id_array_validator_factory('mentioned_role_ids', Role)

# mentioned_users

def parse_mentioned_users(data, guild_id = 0):
    """
    Parses out the mentioned users from the given message data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's id where the message was created at.
    
    Returns
    -------
    mentioned_users : `None`, `tuple` of ``ClientUserBase``
    """
    user_mention_datas = data.get('mentions', None)
    if (user_mention_datas is None) or (not user_mention_datas):
        return
    
    return tuple(sorted(
        (
            User.from_data(user_mention_data, user_mention_data.get('member', None), guild_id)
            for user_mention_data in user_mention_datas
        ),
        key = id_sort_key,
    ))


def put_mentioned_users(mentioned_users, data, defaults, *, guild_id = 0):
    """
    Puts the message's mentioned users' data into the given `data` json serializable object.
    
    Parameters
    ----------
    mentioned_users : `None`, `tuple` of ``ClientUserBase``
        The mentioned users.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (mentioned_users is not None):
        user_mention_datas = []
        
        if (mentioned_users is not None):
            for user in mentioned_users:
                user_data = user.to_data(defaults = defaults, include_internals = True)
                
                if guild_id:
                    try:
                        guild_profile = user.guild_profiles[guild_id]
                    except KeyError:
                        pass
                    else:
                        user_data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
                
                user_mention_datas.append(user_data)
            
        data['mentions'] = user_mention_datas
    
    return data


validate_mentioned_users = nullable_entity_array_validator_factory(
    'mentioned_users', ClientUserBase, sort_key = id_sort_key,
)

# message_id

def parse_message_id(data):
    """
    Parses out a message id from the given data. Not like `parse_id` it first checks for `message_id`
    and then for `id` keys.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    
    Returns
    -------
    message_id : `int`
    """
    try:
        message_id = data['message_id']
    except KeyError:
        pass
    else:
        if message_id is None:
            return 0
        
        return int(message_id)
    
    try:
        message_id = data['id']
    except KeyError:
        pass
    else:
        if message_id is None:
            return 0
        
        return int(message_id)
    
    return 0

put_message_id = entity_id_putter_factory('message_id')

# nonce

parse_nonce = nullable_string_parser_factory('nonce')
put_nonce = url_optional_putter_factory('nonce')
validate_nonce = nullable_string_validator_factory('nonce', 0, NONCE_LENGTH_MAX)

# pinned

parse_pinned = bool_parser_factory('pinned', False)
put_pinned = bool_optional_putter_factory('pinned', False)
validate_pinned = bool_validator_factory('pinned', False)


# poll

def parse_poll(data, old_poll = None):
    """
    Parses the message's polls.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    old_poll : `None | Poll` = `None`, Optional
        The old poll of the message.
    
    Returns
    -------
    poll : `Non | Poll`
    """
    poll_data = data.get('poll', None)
    if poll_data is None:
        poll = None
    else:
        if old_poll is None:
            poll = Poll.from_data(poll_data)
        else:
            poll = old_poll
            old_poll._update_attributes(poll_data)
            
    return poll


def parse_poll_and_change(data, old_poll):
    """
    Parses the poll and returns the difference.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    old_poll : `None | Poll` = `None`, Optional
        The old poll of the message.
    
    Returns
    -------
    poll : `Non | Poll`
    change : `None | PollChange`
    """
    poll_data = data.get('poll', None)
    if poll_data is None:
        if old_poll is None:
            poll = None
            change = None
        else:
            poll = None
            change = PollChange.from_fields(None, None, old_poll)
    else:
        if old_poll is None:
            poll = Poll.from_data(poll_data)
            change = PollChange.from_fields(poll, None, None)
        
        else:
            poll = old_poll
            old_attributes = poll._difference_update_attributes(poll_data)
            if old_attributes:
                change = PollChange.from_fields(None, PollUpdate.from_fields(poll, old_attributes), None)
            else:
                change = None
    
    return poll, change


put_poll = nullable_entity_optional_putter_factory('poll', Poll)
validate_poll = nullable_entity_validator_factory('poll', Poll)

# reactions

def parse_reactions(data, old_reactions = None):
    """
    Parses the message's reactions.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    old_reactions : `None`, ``ReactionMapping`` = `None`, Optional
        The old reactions on the message.
    
    Returns
    -------
    reactions : `None`, ``ReactionMapping``
    """
    reactions_data = data.get('reactions', None)
    if (reactions_data is None) or (not reactions_data):
        new_reactions = None
    else:
        new_reactions = ReactionMapping.from_data(reactions_data)
    
    return merge_update_reaction_mapping(new_reactions, old_reactions)


def put_reactions(reactions, data, defaults):
    """
    Puts the message's reactions' data into the given `data` json serializable object.
    
    Parameters
    ----------
    reactions : `None`, ``ReactionMapping``
        The message's reactions.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (reactions is not None):
        data['reactions'] = reactions.to_data()
    elif defaults:
        data['reactions'] = []
    
    return data


def validate_reactions(reactions):
    """
    Validates message's reactions.
    
    Parameters
    ----------
    reactions : `None`, ``ReactionMapping`` (or compatible)
        Reactions value to validate.
    
    Returns
    -------
    reactions : `None`, ``ReactionMapping``
    
    Raises
    ------
    TypeError
        - `reactions` of invalid type given.
    """
    if reactions is None:
        return None
    
    if isinstance(reactions, ReactionMapping):
        return reactions
    
    return ReactionMapping(lines = reactions)

# referenced_message

def parse_referenced_message(data):
    """
    Parses out the referenced message from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    
    Returns
    -------
    referenced_message : ``None | Message``
    """
    referenced_message_data = data.get('referenced_message', None)
    if (referenced_message_data is not None):
        return Message.from_data(referenced_message_data)
    
    referenced_message_data = data.get('message_reference', None)
    if (referenced_message_data is not None):
        return Message._create_from_partial_data(referenced_message_data)


def put_referenced_message_into(
    referenced_message, data, defaults, *, recursive = True, message_type = MessageType.default
):
    """
    Puts the referenced message's data into the  given `data` json serializable object.
    
    Parameters
    ----------
    referenced_message : ``None | Message``
        The referenced message.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    recursive : `bool` = `True`, Optional (Keyword only)
        Whether the referenced message's data should be included as well.
    
    message_type : ``MessageType` = `MessageType.default`, Optional (Keyword only)
        The message's type. Used to check whether the referenced message should be recursively included.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (referenced_message is not None):
        data['message_reference'] = referenced_message.to_message_reference_data()
        
        if recursive and message_type in (MessageType.inline_reply, MessageType.thread_started):
            data['referenced_message'] = referenced_message.to_data(
                defaults = defaults, include_internals = True, recursive = False
            )
    
    return data


validate_referenced_message = nullable_entity_validator_factory(
    'referenced_message', NotImplemented, include = 'Message'
)

# resolved

def parse_resolved(data, guild_id = 0):
    """
    Parsers out a resolved value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    guild_id : `int` = `0`, Optional
        The respective guild's identifier.
    
    Returns
    -------
    resolved : ``None | Resolved``
    """
    resolved_data = data.get('resolved', None)
    if (resolved_data is not None) and resolved_data:
        return Resolved.from_data(resolved_data, guild_id)


def put_resolved(resolved, data, defaults, *, guild_id = 0):
    """
    Serialises the given resolved value.
    
    Parameters
    ----------
    resolved  : ``None | Resolved``
        The instance to serialise.
        
    data : `dict<str, object>`
        Data to parse from.
    
    defaults : `bool`
        Whether default field values should be included as well.
    
    guild_id : `int` = `None`, Optional (Keyword only)
        The respective guild's identifier to use for handing user guild profiles.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    while True:
        if (resolved is None):
            if not defaults:
                break
            
            resolved_data = {}
        
        else:
            resolved_data = resolved.to_data(defaults = defaults, guild_id = guild_id)
        
        data['resolved'] = resolved_data
        break
    
    return data


validate_resolved = nullable_entity_validator_factory('resolved', NotImplemented, include = 'Resolved')


# role_subscription

parse_role_subscription = nullable_entity_parser_factory('role_subscription_data', MessageRoleSubscription)
put_role_subscription = nullable_entity_optional_putter_factory(
    'role_subscription_data', MessageRoleSubscription
)
validate_role_subscription = nullable_entity_validator_factory('role_subscription', MessageRoleSubscription)


# shared_client_theme

parse_shared_client_theme = nullable_entity_parser_factory('shared_client_theme', SharedClientTheme)
put_shared_client_theme = nullable_entity_optional_putter_factory(
    'shared_client_theme', SharedClientTheme
)
validate_shared_client_theme = nullable_entity_validator_factory('shared_client_theme', SharedClientTheme)


# snapshots

def parse_snapshots(data, guild_id = 0):
    """
    Parsers out the snapshots from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    guild_id : `int` = `0`, Optional
        The respective guild's identifier.
    
    Returns
    -------
    snapshots : `None | tuple<MessageSnapshot>`
    """
    snapshot_datas = data.get('message_snapshots', None)
    if (snapshot_datas is not None) and snapshot_datas:
        return (*(MessageSnapshot.from_data(snapshot_data, guild_id) for snapshot_data in snapshot_datas),)


def put_snapshots(snapshots, data, defaults, *, guild_id = 0):
    """
    Serailises the given snapshots,
    
    Parameters
    ----------
    snapshots : `None | tuple<MessageSnapshot>`
        The instances to serialise.
        
    data : `dict<str, object>`
        Interaction metadata data.
    
    defaults : `bool`
        Whether default field values should be included as well.
    
    guild_id : `int` = `None`, Optional (Keyword only)
        The respective guild's identifier to use for handing user guild profiles.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (snapshots is not None):
        if snapshots is None:
            snapshot_datas = []
        else:
            snapshot_datas = [snapshot.to_data(defaults = defaults, guild_id = guild_id) for snapshot in snapshots]
        
        data['message_snapshots'] = snapshot_datas
    
    return data


validate_snapshots = nullable_object_array_validator_factory('snapshots', NotImplemented, include = 'MessageSnapshot')

# soundboard_sounds

parse_soundboard_sounds = nullable_entity_array_parser_factory('soundboard_sounds', SoundboardSound)
put_soundboard_sounds = nullable_entity_array_optional_putter_factory(
    'soundboard_sounds', SoundboardSound, force_include_internals = True
)
validate_soundboard_sounds = nullable_entity_array_validator_factory('soundboard_sounds', SoundboardSound)

# stickers

parse_stickers = nullable_functional_array_parser_factory(
    'sticker_items', create_partial_sticker_from_partial_data, do_sort = False
)
put_stickers = nullable_functional_array_optional_putter_factory('sticker_items', create_partial_sticker_data)
validate_stickers = nullable_entity_array_validator_factory(
    'stickers', Sticker, sort_key = id_sort_key,
)

# thread

def parse_thread(data, guild_id = 0):
    """
    Parses out the thread started by the message from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Message data.
    guild_id : `int` = `0`, Optional
        The guild's identifier where the message was sent to.
    """
    thread_data = data.get('thread', None)
    if (thread_data is not None):
        return Channel.from_data(thread_data, guild_id = guild_id)


put_thread = nullable_entity_optional_putter_factory(
    'thread', NotImplemented, force_include_internals = True
)
validate_thread = nullable_entity_validator_factory('thread', NotImplemented, include = 'Channel')

# tts

parse_tts = bool_parser_factory('tts', False)
put_tts = bool_optional_putter_factory('tts', False)
validate_tts = bool_validator_factory('tts', False)

# type

parse_type = preinstanced_parser_factory('type', MessageType, MessageType.default)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', MessageType)



__all__ = ()

from itertools import islice

from scarletio import Task, TaskGroup, copy_docs, sleep, to_json

from ..core import KOKORO

from .client_base import DiscordGatewayClientBase
from .client_shard import DiscordGatewayClientShard
from .heartbeat import LATENCY_DEFAULT


def _create_gateways(client, shard_count, gateways):
    """
    Creates the amount of gateways requested.
    
    Parameters
    ----------
    client : ``Client``
        The client to create the gateways for.
    shard_count : `int`
        The amount of gateways to return.
        Should be greater than `1` and must be greater than `0`.
    gateways : `None | tuple<DiscordGatewayClientShard>`
        Already existing gateways to use.
    
    Returns
    -------
    gateways : `tuple<DiscordGatewayClientShard>`
    """
    # Process input
    if gateways is None:
        gateways = []
    else:
        gateways = [*gateways]
    
    gateway_count = len(gateways)
    
    # Adjust gateway count
    if shard_count == gateway_count:
        # nothing to do, should not happen
        pass
    
    elif shard_count > gateway_count:
        # Add new gateways
        gateways.extend(DiscordGatewayClientShard(client, shard_id) for shard_id in range(gateway_count, shard_count))
    
    else:
        # Remove extra gateways
        for _ in range(shard_count, gateway_count):
            gateways.pop().abort()
    
    # return
    return tuple(gateways)


async def _close_and_abort_gateway(gateway):
    """
    Closes and aborts the gateway.
    
    This function is a coroutine.
    
    Parameters
    ----------
    gateway : ``DiscordGatewayClientShard``
        The gateway to stop.
    """
    try:
        await gateway.close()
    finally:
        gateway.abort()


async def _connect_gateway_batch(task_group, gateways):
    """
    Connects a batch of gateways.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        Task group to use for waiters and the runner tasks.
    gateways : `iterable<DiscordGatewayClientShard>`
        The gateways to run.
    
    Returns
    -------
    success : `bool`
    """
    batch_size = 0
    
    # Add gateways
    for gateway in gateways:
        future = task_group.create_future()
        task_group.create_task(gateway.run(future))
        
        batch_size += 1
    
    # Wait for connection waiters. These always have result of `True` on success. While `False` on failure.
    # If a `.runner` ends it also results in `False`. Like this we do not need to differentiate between them..
    while batch_size:
        done_future = await task_group.wait_first_and_pop()
        
        result = done_future.get_result()
        if not result:
            return False
        
        batch_size -= 1
        continue
    
    return True


async def _connect_gateways(task_group, gateways, max_concurrency):
    """
    Connects the given gateways adding each runner's into `task_group`.
    Connects them in batches of `max_concurrency`. If there is more, we wait for rate limits and continue.
    
    This function is a coroutine.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        Task group to use for waiters and the runner tasks.
    gateways : `list<DiscordGatewayClientShard>`
        The gateways to run.
    max_concurrency : `int`
        The maximal amount of shards that can be launched at the same time.
    
    Returns
    -------
    success : `bool`
    """
    chunk_start = 0
    limit = len(gateways)
    # At every step we add up to `max_concurrency` gateways to launch up. When a gateway is launched up
    while True:
        chunk_end = min(chunk_start + max_concurrency, limit)
        
        result = await _connect_gateway_batch(task_group, islice(gateways, chunk_start, chunk_end))
        if not result:
            return False
            
        chunk_start = chunk_end
        if chunk_start >= limit:
            break
        
        # Each gateway does an `IDENTIFY` on connection. You can send `max_concurrency` amount of identifies every
        # second. That means here we sleep `5` seconds. We could rush this 5 seconds, but no need.
        await sleep(5.0, KOKORO)
        continue
    
    return True


class DiscordGatewayClientSharder(DiscordGatewayClientBase):
    """
    Gateway of a client controlling multiple shards.
    
    Attributes
    ----------
    client : ``Client``
        The owner client of the gateway.
    gateways : `tuple<DiscordGatewayClientShard>`
        The controlled gateways.
    """
    __slots__ = ('client', 'gateways')
    
    def __new__(cls, client, shard_count, gateways):
        """
        Creates a sharder gateway with it's default attributes.
        
        Parameters
        ----------
        client : ``Client``
            The owner client of the gateway.
        shard_count : `int`
            The amount of shards to create the gateway with.
        gateways : `None | tuple<DiscordGatewayClientShard>`
            Existing gateways to use.
        """
        gateways = _create_gateways(client, shard_count, gateways)
        
        self = object.__new__(cls)
        self.client = client
        self.gateways = gateways
        return self
    
    
    @copy_docs(DiscordGatewayClientBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(' client = ')
        repr_parts.append(repr(self.client.full_name))
        
        repr_parts.append(', shard_count = ')
        repr_parts.append(repr(len(self.gateways)))
    
    
    @copy_docs(DiscordGatewayClientBase.run)
    async def run(self):
        task_group = TaskGroup(KOKORO)
        
        try:
            result = await _connect_gateways(task_group, self.gateways, self.client._gateway_max_concurrency)
            
            if result:
                # If all shards successfully connected we wait till the first is cancelled.
                # Then we cancel the rest as well.
                finished_task = await task_group.wait_first()
                await self.close()
                finished_task.get_result()
            
            else:
                # If not all shards connected, we stop all of our shards
                await self.close()
        
        except:
            self.abort()
            raise
        
        finally:
            task_group.cancel_all()
        
        return False
    
    
    @property
    @copy_docs(DiscordGatewayClientBase.latency)
    def latency(self):
        total = 0.0
        count = 0
        for gateway in self.gateways:
            kokoro = gateway.kokoro
            if kokoro is None:
                continue
            
            total += kokoro.latency
            count += 1
        
        if count:
            return total / count
        
        return LATENCY_DEFAULT
    
    
    @copy_docs(DiscordGatewayClientBase.terminate)
    async def terminate(self):
        await TaskGroup(KOKORO, (Task(KOKORO, gateway.terminate()) for gateway in self.gateways)).wait_all()
    
    
    @copy_docs(DiscordGatewayClientBase.close)
    async def close(self):
        await TaskGroup(
            KOKORO,
            (Task(KOKORO, _close_and_abort_gateway(gateway)) for gateway in self.gateways),
        ).wait_all()
    
    
    @copy_docs(DiscordGatewayClientBase.abort)
    def abort(self):
        for gateway in self.gateways:
            gateway.abort()
    
    
    @copy_docs(DiscordGatewayClientBase.send_as_json)
    async def send_as_json(self, data):
        data = to_json(data)
        
        task_group = TaskGroup(KOKORO, (Task(KOKORO, gateway._send_json(data)) for gateway in self.gateways))
        failed_task = await task_group.wait_exception()
        if (failed_task is not None):
            task_group.cancel_all()
            failed_task.get_result()
    
    
    @copy_docs(DiscordGatewayClientBase.change_voice_state)
    async def change_voice_state(self, guild_id, channel_id, *, self_deaf = False, self_mute = False):
        gateway = self.get_gateway(guild_id)
        await gateway.change_voice_state(guild_id, channel_id, self_mute = self_mute, self_deaf = self_deaf)
    
    
    @copy_docs(DiscordGatewayClientBase.get_gateway)
    def get_gateway(self, guild_id):
        gateways = self.gateways
        if guild_id:
            gateway = gateways[(guild_id >> 22) % len(gateways)]
        else:
            gateway = gateways[0]
        
        return gateway



__all__ = ()

from ...field_parsers import bool_parser_factory, nullable_string_parser_factory
from ...field_putters import bool_optional_putter_factory, nullable_string_putter_factory
from ...field_validators import bool_validator_factory

from .constants import NAME_LENGTH_MAX, VALUE_LENGTH_MAX


# inline

parse_inline = bool_parser_factory('inline', False)
put_inline = bool_optional_putter_factory('inline', False)
validate_inline = bool_validator_factory('inline', False)


# name

parse_name = nullable_string_parser_factory('name')
put_name = nullable_string_putter_factory('name')


def validate_name(name):
    """
    Validates the given embed author name.
    
    Parameters
    ----------
    name : `None`, `str`, `object`
        Embed author name.
    
    Returns
    -------
    name : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `name`'s type is incorrect.
    ValueError
        - If `name`'s length is out of the expected range.
    """
    if name is None:
        return None
    
    if not isinstance(name, str):
        name = str(name)
    
    name_length = len(name)
    if name_length == 0:
        return None
    
    if name_length > NAME_LENGTH_MAX:
        raise ValueError(
            f'`name` length` must be <= {NAME_LENGTH_MAX}, got {name_length}; name = {name!r}.'
        )
    
    return name

# value

parse_value = nullable_string_parser_factory('value')
put_value = nullable_string_putter_factory('value')


def validate_value(value):
    """
    Validates the given embed author value.
    
    Parameters
    ----------
    value : `None`, `str`, `object`
        Embed author value.
    
    Returns
    -------
    value : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `value`'s type is incorrect.
    ValueError
        - If `value`'s length is out of the expected range.
    """
    if value is None:
        return None
    
    if not isinstance(value, str):
        value = str(value)
    
    value_length = len(value)
    if value_length == 0:
        return None
    
    if value_length > VALUE_LENGTH_MAX:
        raise ValueError(
            f'`value` length` must be <= {VALUE_LENGTH_MAX}, got {value_length}; value = {value!r}.'
        )
    
    return value



__all__ = ('ActivityMetadataHanging',)

from scarletio import copy_docs

from ...utils import DATETIME_FORMAT_CODE

from .base import ActivityMetadataBase, _pop_empty_name
from .fields import (
    parse_created_at, parse_details, parse_emoji, parse_hang_type, put_created_at, put_details,
    put_emoji, put_hang_type, validate_created_at, validate_details, validate_emoji, validate_hang_type
)
from .preinstanced import HangType


class ActivityMetadataHanging(ActivityMetadataBase):
    """
    Represents a Discord hang activity.
    
    Attributes
    ----------
    created_at : `None | DateTime`
        When the status was created.
    
    details : `None | str`
        The activity's text next to the emoji.
    
    emoji : ``None | Emoji``
        The emoji of the activity.
    
    hang_type : ``HangType``
        How the user is hanging.
    """
    __slots__ = ('created_at', 'details', 'emoji', 'hang_type', )
    
    
    def __new__(cls, *, created_at = ..., details = ..., emoji = ..., hang_type = ...):
        """
        Creates a new hang activity metadata.
        
        Parameters
        ----------
        created_at : `None | DateTime`, Optional (Keyword only)
            When the activity was created.
        
        emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji of the activity.
        
        details : `None`, `str`, Optional (Keyword only)
            The activity's text under it's emoji.
        
        hang_type : ``HangType``, Optional (Keyword only)
            How the user is hanging.
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        # created_at
        if created_at is ...:
            created_at = None
        else:
            created_at = validate_created_at(created_at)
        
        # details
        if details is ...:
            details = None
        else:
            details = validate_details(details)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # hang_type
        if hang_type is ...:
            hang_type = HangType.custom
        else:
            hang_type = validate_hang_type(hang_type)
        
        # Construct
        self = object.__new__(cls)
        self.created_at = created_at
        self.details = details
        self.emoji = emoji
        self.hang_type = hang_type
        return self
        
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        _pop_empty_name(keyword_parameters)
        return cls(
            created_at = keyword_parameters.pop('created_at', ...),
            details = keyword_parameters.pop('details', ...),
            emoji = keyword_parameters.pop('emoji', ...),
            hang_type = keyword_parameters.pop('hang_type', ...),
        )
    
    
    @copy_docs(ActivityMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # created_at
        created_at = self.created_at
        if (created_at is not None):
            repr_parts.append(' created_at = ')
            repr_parts.append(format(created_at, DATETIME_FORMAT_CODE))
            
            field_added = True
        else:
            field_added = False
        
        # details
        details = self.details
        if (details is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = False
            
            repr_parts.append(' details = ')
            repr_parts.append(repr(details))
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' emoji = ')
            repr_parts.append(repr(emoji))
        
        # hang_type
        hang_type = self.hang_type
        if (hang_type is not HangType.none):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' hang_type = ')
            repr_parts.append(hang_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(hang_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # created_a
        created_at = self.created_at
        if (created_at is not None):
            hash_value ^= hash(created_at)
        
        # details
        details = self.details
        if (details is not None):
            hash_value ^= hash(details)
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # hang_type
        hang_type = self.hang_type
        if (hang_type is not HangType.none):
            hash_value ^= hash(hang_type)
        
        return hash_value
    
    
    @copy_docs(ActivityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # details
        if self.details != other.details:
            return False
        
        # emoji
        if self.emoji != other.emoji:
            return False
        
        # hang_type
        if self.hang_type is not other.hang_type:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self._update_attributes(data)
        return self
    
    
    @copy_docs(ActivityMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False, user = False):
        data = {}
        
        put_hang_type(self.hang_type, data, defaults)
        
        if user or include_internals:
            put_details(self.details, data, defaults)
            put_emoji(self.emoji, data, defaults)
        
        if include_internals:
            data['name'] = 'Hang Status'
            put_created_at(self.created_at, data, defaults)
        
        return data
    
    
    @copy_docs(ActivityMetadataBase._update_attributes)
    def _update_attributes(self, data):
        self.created_at = parse_created_at(data)
        self.details = parse_details(data)
        self.emoji = parse_emoji(data)
        self.hang_type = parse_hang_type(data)
    
    
    @copy_docs(ActivityMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = {}
        
        # created_at
        created_at = parse_created_at(data)
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        # details
        details = parse_details(data)
        if self.details != details:
            old_attributes['details'] = self.details
            self.details = details
        
        # emoji
        emoji = parse_emoji(data)
        if self.emoji != emoji:
            old_attributes['emoji'] = self.emoji
            self.emoji = emoji
        
        # hang_type
        hang_type = parse_hang_type(data)
        if self.hang_type is not hang_type:
            old_attributes['hang_type'] = self.hang_type
            self.hang_type = hang_type
        
        return old_attributes
    
    
    @copy_docs(ActivityMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.created_at = self.created_at
        new.emoji = self.emoji
        new.details = self.details
        new.hang_type = self.hang_type
        return new
    
    
    def copy_with(self, *, created_at = ..., details = ..., emoji = ..., hang_type = ...):
        """
        Copies the hang activity metadata with the given fields.
        
        Parameters
        ----------
        created_at : `None | DateTime`, Optional (Keyword only)
            When the status was created.
        
        details : `None`, `str`, Optional (Keyword only)
            The activity's next to its emoji.
        
        emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji of the activity.
        
        hang_type : ``HangType``, Optional (Keyword only)
            How the user is hanging.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        # created_at
        if created_at is ...:
            created_at = self.created_at
        else:
            created_at = validate_created_at(created_at)
        
        # details
        if details is ...:
            details = self.details
        else:
            details = validate_details(details)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # hang_type
        if hang_type is ...:
            hang_type = self.hang_type
        else:
            hang_type = validate_hang_type(hang_type)
        
        # Construct
        new = object.__new__(type(self))
        new.created_at = created_at
        new.emoji = emoji
        new.details = details
        new.hang_type = hang_type
        return new
    
    
    @copy_docs(ActivityMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        _pop_empty_name(keyword_parameters)
        return self.copy_with(
            created_at = keyword_parameters.pop('created_at', ...),
            details = keyword_parameters.pop('details', ...),
            emoji = keyword_parameters.pop('emoji', ...),
            hang_type = keyword_parameters.pop('hang_type', ...),
        )
    
    
    @property
    def name(self):
        """
        Returns the activity's display text.
        
        Returns
        -------
        name : `str`
        """
        return self.hang_type.name_getter(self)



__all__ = ('ApplicationRoleConnection',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_metadata_values, parse_platform_name, parse_platform_user_name, put_metadata_values,
    put_platform_name, put_platform_user_name, validate_metadata_values, validate_platform_name,
    validate_platform_user_name
)


class ApplicationRoleConnection(RichAttributeErrorBaseType):
    """
    Application role connection attached to a user.
    
    Attributes
    ----------
    platform_name : `None`, `str`
        The vanity name of the platform the application represents.
    platform_user_name : `None`, `str`
        The name of the user on the application's platform.
    metadata_values : `None | dict<str, str>`
        Metadata key to attached value relation.
    """
    __slots__ = ('platform_name', 'platform_user_name', 'metadata_values')
    
    def __new__(cls, *, platform_name = ..., platform_user_name = ..., metadata_values = ...):
        """
        Creates a new application role connection.
        
        Parameters
        ----------
        platform_name : `None`, `str`, Optional (Keyword only)
            The vanity name of the platform the application represents.
        platform_user_name : `None`, `str`, Optional (Keyword only)
            The name of the user on the application's platform.
        metadata_values : `None | dict<str, str>`, Optional (Keyword only)
            Metadata key to attached value relation.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # platform_name
        if platform_name is ...:
            platform_name = None
        else:
            platform_name = validate_platform_name(platform_name)
        
        # platform_user_name
        if platform_user_name is ...:
            platform_user_name = None
        else:
            platform_user_name = validate_platform_user_name(platform_user_name)
        
        # metadata_values
        if metadata_values is ...:
            metadata_values = None
        else:
            metadata_values = validate_metadata_values(metadata_values)
        
        self = object.__new__(cls)
        self.platform_name = platform_name
        self.platform_user_name = platform_user_name
        self.metadata_values = metadata_values
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an application role connection from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Role connection data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.platform_name = parse_platform_name(data)
        self.platform_user_name = parse_platform_user_name(data)
        self.metadata_values = parse_metadata_values(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application role connection to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_platform_name(self.platform_name, data, defaults)
        put_platform_user_name(self.platform_user_name, data, defaults)
        put_metadata_values(self.metadata_values, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the role connection's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        platform_name = self.platform_name
        if (platform_name is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            repr_parts.append(' platform_name = ')
            repr_parts.append(repr(platform_name))
        
        platform_user_name = self.platform_user_name
        if (platform_user_name is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            repr_parts.append(' platform_user_name = ')
            repr_parts.append(repr(platform_user_name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the hash value of the role connection."""
        hash_value = 0
        
        # platform_name
        platform_name = self.platform_name
        if (platform_name is not None):
            hash_value ^= hash(platform_name)
        
        # platform_user_name
        platform_user_name = self.platform_user_name
        if (platform_user_name is not None):
            hash_value ^= hash(platform_user_name)
        
        # metadata_values
        metadata_values = self.metadata_values
        if (metadata_values is not None):
            hash_value ^= len(metadata_values)
            
            for field_name, field_value in metadata_values.items():
                hash_value ^= hash(field_name) ^ hash(field_value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two role connections are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # platform_name
        if self.platform_name != other.platform_name:
            return False
        
        # platform_user_name
        if self.platform_user_name != other.platform_user_name:
            return False
        
        # metadata_values
        if self.metadata_values != other.metadata_values:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the application role connection.
        
        Returns
        -------
        new : `instance<type<cls>>`
        """
        new = object.__new__(type(self))
        new.platform_name = self.platform_name
        new.platform_user_name = self.platform_user_name
        metadata_values = self.metadata_values
        if (metadata_values is not None):
            metadata_values = metadata_values.copy()
        new.metadata_values = metadata_values
        return new
    
    
    def copy_with(self, *, platform_name = ..., platform_user_name = ..., metadata_values = ...):
        
        """
        Copies the application role connection with the given fields.
        
        Parameters
        ----------
        platform_name : `None`, `str`, Optional (Keyword only)
            The vanity name of the platform the application represents.
        platform_user_name : `None`, `str`, Optional (Keyword only)
            The name of the user on the application's platform.
        metadata_values : `None | dict<str, str>`, Optional (Keyword only)
            Metadata key to attached value relation.
        
        Returns
        -------
        new : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # platform_name
        if platform_name is ...:
            platform_name = self.platform_name
        else:
            platform_name = validate_platform_name(platform_name)
        
        # platform_user_name
        if platform_user_name is ...:
            platform_user_name = self.platform_user_name
        else:
            platform_user_name = validate_platform_user_name(platform_user_name)
        
        # metadata_values
        if metadata_values is ...:
            metadata_values = self.metadata_values
            if (metadata_values is not None):
                metadata_values = metadata_values.copy()
        else:
            metadata_values = validate_metadata_values(metadata_values)
        
        new = object.__new__(type(self))
        new.platform_name = platform_name
        new.platform_user_name = platform_user_name
        new.metadata_values = metadata_values
        return new
    
    
    def translate_value(self, metadata):
        """
        Translates the given metadata's respective value ``.metadata_values`` to their type-correct representation.
        Returns `None` on failure.
        
        Parameters
        ----------
        metadata : ``ApplicationRoleConnectionMetadata``
            Application role connection metadata to translate with.
        
        Returns
        -------
        translated_value : `None | object`
        """
        metadata_values = self.metadata_values
        if (metadata_values is None):
            return
        
        try:
            raw_value = metadata_values[metadata.key]
        except KeyError:
            return
        
        return metadata.type.value_type.deserializer(raw_value)
    
    
    def translate_values(self, metadatas):
        """
        Translates the ``.metadata_values`` of the role connection to their type-correct representation.
        
        Parameters
        ----------
        metadatas : `iterable` of ``ApplicationRoleConnectionMetadata``
            Application role connection metadatas to translate with.
        
        Returns
        -------
        translated_values : `dict` of (``ApplicationRoleConnectionMetadata``, `object`) items
        """
        translated_values = {}
        
        metadata_values = self.metadata_values
        if (metadata_values is not None):
            for metadata in metadatas:
                try:
                    raw_value = metadata_values[metadata.key]
                except KeyError:
                    continue
                
                converted_value = metadata.type.value_type.deserializer(raw_value)
                if (converted_value is not None):
                    translated_values[metadata] = converted_value
        
        return translated_values



__all__ = ()


from ...application_command import ApplicationCommand
from ...auto_moderation import AutoModerationRule
from ...channel import Channel
from ...field_validators import entity_id_validator_factory
from ...guild import Guild
from ...integration import Integration
from ...scheduled_event import ScheduledEvent
from ...user import ClientUserBase, User
from ...webhook import Webhook

from ..audit_log_entry import AuditLogEntry


# application_commands

def parse_application_commands(data):
    """
    Parses an audit log entry's application commands.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    application_commands : `None | dict<int, ApplicationCommand>`
    """
    try:
        application_command_datas = data['application_commands']
    except KeyError:
        return None
    
    if not application_command_datas:
        return None
    
    application_commands = {}
    
    for application_command_data in application_command_datas:
        application_command = ApplicationCommand.from_data(application_command_data)
        application_commands[application_command.id] = application_command
    
    return application_commands


def put_application_commands(application_commands, data, defaults):
    """
    Serialises the application commands into the given data.
    
    Parameters
    ----------
    application_commands : `None | dict<int, ApplicationCommand>`
        The application commands to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (application_commands is not None):
        if application_commands is None:
            application_command_datas = []
        else:
            application_command_datas = [
                application_command.to_data(defaults = defaults, include_internals = True)
                for application_command in application_commands.values()
            ]
        data['application_commands'] = application_command_datas
    
    return data


def validate_application_commands(application_commands):
    """
    Validates the given application commands.
    
    Parameters
    ----------
    application_commands : `None | iterable<ApplicationCommand>`
        The application commands to validate.
    
    Returns
    -------
    application_commands : `None | dict<int, ApplicationCommand>`
    
    Raises
    ------
    TypeError
        - If `application_commands`'s type is incorrect.
    """
    validated_application_commands = None
    
    if application_commands is None:
        return validated_application_commands
        
    if (getattr(application_commands, '__iter__', None) is None):
        raise TypeError(
            f'`application_commands` can be `None`, `iterable` of `{ApplicationCommand.__name__}` elements, '
            f'got {type(application_commands).__name__}; {application_commands!r}.'
        )
    
    for application_command in application_commands:
        if not isinstance(application_command, ApplicationCommand):
            raise TypeError(
                f'`application_commands` elements can be `{ApplicationCommand.__name__}`, '
                f'got {type(application_command).__name__}; {application_command!r}; '
                f'application_commands = {application_commands!r}.'
            )
        
        if validated_application_commands is None:
            validated_application_commands = {}
        
        validated_application_commands[application_command.id] = application_command
    
    return validated_application_commands


# auto_moderation_rules

def parse_auto_moderation_rules(data):
    """
    Parses an audit log entry's application commands.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    auto_moderation_rules : `None | dict<int, AutoModerationRule>`
    """
    try:
        auto_moderation_rule_datas = data['auto_moderation_rules']
    except KeyError:
        return None
    
    if not auto_moderation_rule_datas:
        return None
    
    auto_moderation_rules = {}
    
    for auto_moderation_rule_data in auto_moderation_rule_datas:
        auto_moderation_rule = AutoModerationRule.from_data(auto_moderation_rule_data)
        auto_moderation_rules[auto_moderation_rule.id] = auto_moderation_rule
    
    return auto_moderation_rules


def put_auto_moderation_rules(auto_moderation_rules, data, defaults):
    """
    Serialises the auto moderation rules into the given data.
    
    Parameters
    ----------
    auto_moderation_rules : `None | dict<int, AutoModerationRule>`
        The auto moderation rules to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (auto_moderation_rules is not None):
        if auto_moderation_rules is None:
            auto_moderation_rule_datas = []
        else:
            auto_moderation_rule_datas = [
                auto_moderation_rule.to_data(defaults = defaults, include_internals = True)
                for auto_moderation_rule in auto_moderation_rules.values()
            ]
        data['auto_moderation_rules'] = auto_moderation_rule_datas
    
    return data


def validate_auto_moderation_rules(auto_moderation_rules):
    """
    Validates the given auto moderation rules.
    
    Parameters
    ----------
    auto_moderation_rules : `None | iterable<AutoModerationRule>`
        The auto moderation rules to validate.
    
    Returns
    -------
    auto_moderation_rules : `None | dict<int, AutoModerationRule>`
    
    Raises
    ------
    TypeError
        - If `auto_moderation_rules`'s type is incorrect.
    """
    validated_auto_moderation_rules = None
    
    if auto_moderation_rules is None:
        return validated_auto_moderation_rules
        
    if (getattr(auto_moderation_rules, '__iter__', None) is None):
        raise TypeError(
            f'`auto_moderation_rules` can be `None`, `iterable` of `{AutoModerationRule.__name__}` elements, '
            f'got {type(auto_moderation_rules).__name__}; {auto_moderation_rules!r}.'
        )
    
    for auto_moderation_rule in auto_moderation_rules:
        if not isinstance(auto_moderation_rule, AutoModerationRule):
            raise TypeError(
                f'`auto_moderation_rules` elements can be `{AutoModerationRule.__name__}`, '
                f'got {type(auto_moderation_rule).__name__}; {auto_moderation_rule!r}; '
                f'auto_moderation_rules = {auto_moderation_rules!r}.'
            )
        
        if validated_auto_moderation_rules is None:
            validated_auto_moderation_rules = {}
        
        validated_auto_moderation_rules[auto_moderation_rule.id] = auto_moderation_rule
    
    return validated_auto_moderation_rules


# entries

def parse_entries(data, parent = None):
    """
    Parses an audit log entry's entries.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    parent : `None | AuditLog` = `None`, Optional
        The respective parent to backlink to.
    
    Returns
    -------
    entries : `None | list<AuditLogEntry>`
    """
    try:
        entry_datas = data['audit_log_entries']
    except KeyError:
        return None
    
    if not entry_datas:
        return None
    
    entries = None
    
    for entry_data in entry_datas:
        entry = AuditLogEntry.from_data(entry_data, parent)
        if (entry is not None):
            if entries is None:
                entries = []
            
            entries.append(entry)
    
    return entries


def put_entries(entries, data, defaults):
    """
    Serialises the entries into the given data.
    
    Parameters
    ----------
    entries : `None | dict<int, AutoModerationRule>`
        The entries to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (entries is not None):
        if entries is None:
            entry_datas = []
        else:
            entry_datas = [
                entry.to_data(defaults = defaults)
                for entry in entries
            ]
        data['audit_log_entries'] = entry_datas
    
    return data


def validate_entries(entries):
    """
    Validates the given entries.
    
    Parameters
    ----------
    entries : `None | iterable<AuditLogEntry>`
        The entries to validate.
    
    Returns
    -------
    entries : `None | list<AuditLogEntry>`
    
    Raises
    ------
    TypeError
        - If `entries`'s type is incorrect.
    """
    validated_entries = None
    
    if entries is None:
        return validated_entries
        
    if (getattr(entries, '__iter__', None) is None):
        raise TypeError(
            f'`entries` can be `None`, `iterable` of `{AuditLogEntry.__name__}` elements, '
            f'got {type(entries).__name__}; {entries!r}.'
        )
    
    for entry in entries:
        if not isinstance(entry, AuditLogEntry):
            raise TypeError(
                f'`entries` elements can be `{AuditLogEntry.__name__}`, '
                f'got {type(entry).__name__}; {entry!r}; '
                f'entries = {entries!r}.'
            )
        
        if validated_entries is None:
            validated_entries = set()
        
        validated_entries.add(entry)
    
    if (validated_entries is not None):
        return sorted(validated_entries)


# guild_id

validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# integrations

def parse_integrations(data):
    """
    Parses an audit log entry's application commands.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    integrations : `None | dict<int, Integration>`
    """
    try:
        integration_datas = data['integrations']
    except KeyError:
        return None
    
    if not integration_datas:
        return None
    
    integrations = {}
    
    for integration_data in integration_datas:
        integration = Integration.from_data(integration_data)
        integrations[integration.id] = integration
    
    return integrations


def put_integrations(integrations, data, defaults):
    """
    Serialises the integrations into the given data.
    
    Parameters
    ----------
    integrations : `None | dict<int, Integration>`
        The integrations to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (integrations is not None):
        if integrations is None:
            integration_datas = []
        else:
            integration_datas = [
                integration.to_data(defaults = defaults, include_internals = True)
                for integration in integrations.values()
            ]
        data['integrations'] = integration_datas
    
    return data


def validate_integrations(integrations):
    """
    Validates the given integrations.
    
    Parameters
    ----------
    integrations : `None | iterable<Integration>`
        The integrations to validate.
    
    Returns
    -------
    integrations : `None | dict<int, Integration>`
    
    Raises
    ------
    TypeError
        - If `integrations`'s type is incorrect.
    """
    validated_integrations = None
    
    if integrations is None:
        return validated_integrations
        
    if (getattr(integrations, '__iter__', None) is None):
        raise TypeError(
            f'`integrations` can be `None`, `iterable` of `{Integration.__name__}` elements, '
            f'got {type(integrations).__name__}; {integrations!r}.'
        )
    
    for integration in integrations:
        if not isinstance(integration, Integration):
            raise TypeError(
                f'`integrations` elements can be `{Integration.__name__}`, '
                f'got {type(integration).__name__}; {integration!r}; '
                f'integrations = {integrations!r}.'
            )
        
        if validated_integrations is None:
            validated_integrations = {}
        
        validated_integrations[integration.id] = integration
    
    return validated_integrations


# scheduled_events

def parse_scheduled_events(data):
    """
    Parses an audit log entry's application commands.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    scheduled_events : `None | dict<int, ScheduledEvent>`
    """
    try:
        scheduled_event_datas = data['guild_scheduled_events']
    except KeyError:
        return None
    
    if not scheduled_event_datas:
        return None
    
    scheduled_events = {}
    
    for scheduled_event_data in scheduled_event_datas:
        scheduled_event = ScheduledEvent.from_data(scheduled_event_data)
        scheduled_events[scheduled_event.id] = scheduled_event
    
    return scheduled_events


def put_scheduled_events(scheduled_events, data, defaults):
    """
    Serialises the scheduled events into the given data.
    
    Parameters
    ----------
    scheduled_events : `None | dict<int, ScheduledEvent>`
        The scheduled events to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (scheduled_events is not None):
        if scheduled_events is None:
            scheduled_event_datas = []
        else:
            scheduled_event_datas = [
                scheduled_event.to_data(defaults = defaults, include_internals = True)
                for scheduled_event in scheduled_events.values()
            ]
        data['guild_scheduled_events'] = scheduled_event_datas
    
    return data


def validate_scheduled_events(scheduled_events):
    """
    Validates the given scheduled events.
    
    Parameters
    ----------
    scheduled_events : `None | iterable<ScheduledEvent>`
        The scheduled events to validate.
    
    Returns
    -------
    scheduled_events : `None | dict<int, ScheduledEvent>`
    
    Raises
    ------
    TypeError
        - If `scheduled_events`'s type is incorrect.
    """
    validated_scheduled_events = None
    
    if scheduled_events is None:
        return validated_scheduled_events
        
    if (getattr(scheduled_events, '__iter__', None) is None):
        raise TypeError(
            f'`scheduled_events` can be `None`, `iterable` of `{ScheduledEvent.__name__}` elements, '
            f'got {type(scheduled_events).__name__}; {scheduled_events!r}.'
        )
    
    for scheduled_event in scheduled_events:
        if not isinstance(scheduled_event, ScheduledEvent):
            raise TypeError(
                f'`scheduled_events` elements can be `{ScheduledEvent.__name__}`, '
                f'got {type(scheduled_event).__name__}; {scheduled_event!r}; '
                f'scheduled_events = {scheduled_events!r}.'
            )
        
        if validated_scheduled_events is None:
            validated_scheduled_events = {}
        
        validated_scheduled_events[scheduled_event.id] = scheduled_event
    
    return validated_scheduled_events


# threads

def parse_threads(data, guild_id = 0):
    """
    Parses an audit log entry's threads.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    guild_id : `int` = `0`, Optional
        The respective guild's identifier.
    
    Returns
    -------
    threads : `None | dict<int, Channel>`
    """
    try:
        thread_datas = data['threads']
    except KeyError:
        return None
    
    if not thread_datas:
        return None
    
    threads = {}
    
    for thread_data in thread_datas:
        thread = Channel.from_data(thread_data, None, guild_id)
        threads[thread.id] = thread
    
    return threads


def put_threads(threads, data, defaults):
    """
    Serialises the threads into the given data.
    
    Parameters
    ----------
    threads : `None | dict<int, Channel>`
        The threads to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (threads is not None):
        if threads is None:
            thread_datas = []
        else:
            thread_datas = [
                thread.to_data(defaults = defaults, include_internals = True)
                for thread in threads.values()
            ]
        data['threads'] = thread_datas
    
    return data


def validate_threads(threads):
    """
    Validates the given channels.
    
    Parameters
    ----------
    threads : `None | iterable<Channel>`
        The channels to validate.
    
    Returns
    -------
    threads : `None | dict<int, Channel>`
    
    Raises
    ------
    TypeError
        - If `threads`'s type is incorrect.
    """
    validated_threads = None
    
    if threads is None:
        return validated_threads
        
    if (getattr(threads, '__iter__', None) is None):
        raise TypeError(
            f'`threads` can be `None`, `iterable` of `{Channel.__name__}` elements, '
            f'got {type(threads).__name__}; {threads!r}.'
        )
    
    for thread in threads:
        if not isinstance(thread, Channel):
            raise TypeError(
                f'`threads` elements can be `{Channel.__name__}`, '
                f'got {type(thread).__name__}; {thread!r}; '
                f'threads = {threads!r}.'
            )
        
        if validated_threads is None:
            validated_threads = {}
        
        validated_threads[thread.id] = thread
    
    return validated_threads


# users

def parse_users(data):
    """
    Parses an audit log entry's users.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    users : `None | dict<int, ClientUserBase>`
    """
    try:
        user_datas = data['users']
    except KeyError:
        return None
    
    if not user_datas:
        return None
    
    users = {}
    
    for user_data in user_datas:
        user = User.from_data(user_data)
        users[user.id] = user
    
    return users


def put_users(users, data, defaults):
    """
    Serialises the users into the given data.
    
    Parameters
    ----------
    users : `None | dict<int, ClientUserBase>`
        The users to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (users is not None):
        if users is None:
            user_datas = []
        else:
            user_datas = [
                user.to_data(defaults = defaults, include_internals = True)
                for user in users.values()
            ]
        data['users'] = user_datas
    
    return data


def validate_users(users):
    """
    Validates the given users.
    
    Parameters
    ----------
    users : `None | iterable<ClientUserBase>`
        The users to validate.
    
    Returns
    -------
    users : `None | dict<int, ClientUserBase>`
    
    Raises
    ------
    TypeError
        - If `users`'s type is incorrect.
    """
    validated_users = None
    
    if users is None:
        return validated_users
        
    if (getattr(users, '__iter__', None) is None):
        raise TypeError(
            f'`users` can be `None`, `iterable` of `{ClientUserBase.__name__}` elements, '
            f'got {type(users).__name__}; {users!r}.'
        )
    
    for user in users:
        if not isinstance(user, ClientUserBase):
            raise TypeError(
                f'`users` elements can be `{ClientUserBase.__name__}`, '
                f'got {type(user).__name__}; {user!r}; '
                f'users = {users!r}.'
            )
        
        if validated_users is None:
            validated_users = {}
        
        validated_users[user.id] = user
    
    return validated_users


# webhooks

def parse_webhooks(data):
    """
    Parses an audit log entry's webhooks.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    webhooks : `None | dict<int, Webhook>`
    """
    try:
        webhook_datas = data['webhooks']
    except KeyError:
        return None
    
    if not webhook_datas:
        return None
    
    webhooks = {}
    
    for webhook_data in webhook_datas:
        webhook = Webhook.from_data(webhook_data)
        webhooks[webhook.id] = webhook
    
    return webhooks


def put_webhooks(webhooks, data, defaults):
    """
    Serialises the webhooks into the given data.
    
    Parameters
    ----------
    webhooks : `None | dict<int, Webhook>`
        The webhooks to serialise.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (webhooks is not None):
        if webhooks is None:
            webhook_datas = []
        else:
            webhook_datas = [
                webhook.to_data(defaults = defaults, include_internals = True)
                for webhook in webhooks.values()
            ]
        data['webhooks'] = webhook_datas
    
    return data


def validate_webhooks(webhooks):
    """
    Validates the given webhooks.
    
    Parameters
    ----------
    webhooks : `None | iterable<Webhook>`
        The webhooks to validate.
    
    Returns
    -------
    webhooks : `None | dict<int, Webhook>`
    
    Raises
    ------
    TypeError
        - If `webhooks`'s type is incorrect.
    """
    validated_webhooks = None
    
    if webhooks is None:
        return validated_webhooks
        
    if (getattr(webhooks, '__iter__', None) is None):
        raise TypeError(
            f'`webhooks` can be `None`, `iterable` of `{Webhook.__name__}` elements, '
            f'got {type(webhooks).__name__}; {webhooks!r}.'
        )
    
    for webhook in webhooks:
        if not isinstance(webhook, Webhook):
            raise TypeError(
                f'`webhooks` elements can be `{Webhook.__name__}`, '
                f'got {type(webhook).__name__}; {webhook!r}; '
                f'webhooks = {webhooks!r}.'
            )
        
        if validated_webhooks is None:
            validated_webhooks = {}
        
        validated_webhooks[webhook.id] = webhook
    
    return validated_webhooks




__all__ = ('merge_update_reaction_mapping_lines',)


def merge_update_reaction_mapping_lines(new_lines, old_lines):
    """
    Merges reaction mapping lines.
    
    Parameters
    ----------
    new_lines : `None | dict<Reaction, ReactionMappingLine>`
        New lines.
    old_lines : `None | dict<Reaction, ReactionMappingLine>`
        Old lines.
    
    Returns
    -------
    line : `None | list<ReactionMappingLine>`
    """
    if new_lines is None:
        return None
    
    if old_lines is None:
        return new_lines
    
    # Remove not alive answer_id-s
    for reaction, old_line in [*old_lines.items()]:
        if reaction not in new_lines:
            del old_lines[reaction]
    
    # Merge old with new.
    for reaction, new_line in new_lines.items():
        old_line = old_lines.get(reaction, None)
        if old_line is None:
            old_lines[reaction] = new_line
        else:
            old_line._merge_with(new_line)

    return old_lines



__all__ = ('Role', )

from scarletio import copy_docs, export

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...color import Color
from ...core import GUILDS, ROLES
from ...http.urls import build_role_icon_url, build_role_icon_url_as
from ...permission.constants import PERMISSION_KEY
from ...permission.permission import PERMISSION_NONE, Permission
from ...precreate_helpers import process_precreate_parameters, raise_extra
from ...utils import DATETIME_FORMAT_CODE

from ..role_color_configuration import RoleColorConfiguration
from ..role_manager_metadata import RoleManagerMetadataBase

from .constants import ROLE_MANAGER_DEFAULT
from .fields import (
    parse_color, parse_color_configuration, parse_flags, parse_id, parse_manager, parse_mentionable, parse_name,
    parse_permissions, parse_position, parse_separated, parse_unicode_emoji, put_color, put_color_configuration,
    put_flags, put_id, put_manager, put_mentionable, put_name, put_permissions, put_position, put_separated,
    put_unicode_emoji, validate_color, validate_color_configuration, validate_flags, validate_id, validate_manager,
    validate_manager_metadata, validate_manager_type, validate_mentionable, validate_name, validate_permissions,
    validate_position, validate_separated, validate_unicode_emoji, validate_guild_id
)
from .flags import RoleFlag
from .preinstanced import RoleManagerType


ROLE_ICON = IconSlot('icon', 'icon')


PRECREATE_FIELDS = {
    'color': ('color', validate_color),
    'color_configuration': ('color_configuration', validate_color_configuration),
    'flags': ('flags', validate_flags),
    'manager_metadata': ('manager_metadata', validate_manager_metadata),
    'manager_type': ('manager_type', validate_manager_type),
    'mentionable': ('mentionable', validate_mentionable),
    'name': ('name', validate_name),
    'permissions': ('permissions', validate_permissions),
    'position': ('position', validate_position),
    'separated': ('separated', validate_separated),
}


@export
class Role(DiscordEntity, immortal = True):
    """
    Represents a Discord guild's role.
    
    Attributes
    ----------
    color : ``Color``
        The role's color. If the color equals to `Color(0)`, then it is ignored meanwhile calculating towards a
        user's display color.
    
    color_configuration : ``RoleColorConfiguration``
        The role's color's configuration. Its `.color_primary` always equals to ``.color``.
    
    flags : ``RoleFlag``
        The role flags.
    
    icon_hash : `int`
        The guild's icon's hash in `uint128`.
        
        Mutually exclusive with ``.unicode_emoji``
        
    icon_type : ``IconType``
        The guild's icon's type.
        
        Mutually exclusive with ``.unicode_emoji``
    
    guild_id : `int`
        The role's guild's identifier.
    
    id : `int`
        The unique identifier number of the role.
    
    manager_metadata : ``RoleManagerMetadataBase``
        If the role is managed, then it's manager's metadata.
    
    manager_type : ``RoleManagerType``
        If the role is managed, defines by what it is.
    
    mentionable : `bool`
        Whether the role can be mentioned.
    
    name : `str`
        The name of the role. Can be empty string, if the role has no known name.
    
    permissions : ``Permission``
        The permissions of the users having the role.
    
    position : `int`
        The role's position.
    
    separated : `bool`
        Users show up in separated groups by their highest `separated` role.
    
    unicode_emoji : ``None | Emoji``
        Unicode emoji icon of the role.
        
        Mutually exclusive with ``.icon_hash`` and ``.icon_type``.
    """
    __slots__ = (
        'color', 'color_configuration', 'flags', 'guild_id', 'manager_metadata', 'manager_type', 'mentionable', 'name',
        'permissions', 'position', 'separated', 'unicode_emoji'
    )
    
    icon = ROLE_ICON
    
    
    def __new__(
        cls,
        *,
        color = ...,
        color_configuration = ...,
        flags = ...,
        icon = ...,
        manager = ...,
        mentionable = ...,
        name = ...,
        permissions = ...,
        position = ...,
        separated = ...,
        unicode_emoji = ...,
    ):
        """
        Creates a new partial role from the given fields.
        
        Parameters
        ----------
        color : ``None | int | Color`, Optional (Keyword only)
            The role's color.
        
        color_configuration : ``None | RoleColorConfiguration``, Optional (Keyword only)
            The role's color's configuration.
        
        flags : ``None | int | RoleFlag``, Optional (Keyword only)
            The role's flags.
        
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The role's icon.
            
            > Mutually exclusive with the `unicode_emoji` parameter.
        
        manager : ``None | (RoleManagerType, RoleManagerMetadataBase)``, Optional (Keyword only)
            The role's manager.
        
        mentionable : `bool`, Optional (Keyword only)
            Whether the role can be mentioned.
        
        name : `str`, Optional (Keyword only)
            The role's name.
        
        permissions : `int`, ``Permission``, Optional (Keyword only)
            The permissions of the users having the role.
        
        position : `int`, Optional (Keyword only)
            The role's position.
        
        separated : `bool`, Optional (Keyword only)
            Users show up in separated groups by their highest `separated` role.
        
        unicode_emoji : ``None | Emoji``, Optional (Keyword only)
            The role's icon as an unicode emoji.
            
            > Mutually exclusive with the `icon` parameter.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - Value of invalid type given.
        ValueError
            - Value of invalid value given.
            - `icon` and `unicode_emoji` are mutually exclusive.
        """
        # color
        if color is ...:
            color = Color()
        else:
            color = validate_color(color)
        
        # color_configuration
        if color_configuration is ...:
            color_configuration = RoleColorConfiguration.create_empty()
        else:
            color_configuration = validate_color_configuration(color_configuration)
        
        # flags
        if flags is ...:
            flags = RoleFlag()
        else:
            flags = validate_flags(flags)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon, allow_data = True)
        
        # manager
        if manager is ...:
            manager_type = RoleManagerType.none
            manager_metadata = ROLE_MANAGER_DEFAULT
        else:
            manager_type, manager_metadata = validate_manager(manager)
        
        # mentionable
        if mentionable is ...:
            mentionable = False
        else:
            mentionable = validate_mentionable(mentionable)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # permissions
        if permissions is ...:
            permissions = Permission()
        else:
            permissions = validate_permissions(permissions)
        
        # position
        if position is ...:
            position = 0
        else:
            position = validate_position(position)
        
        # separated
        if separated is ...:
            separated = False
        else:
            separated = validate_separated(separated)
        
        # unicode_emoji
        if unicode_emoji is ...:
            unicode_emoji = None
        else:
            unicode_emoji = validate_unicode_emoji(unicode_emoji)
        
        if (icon is not None) and (unicode_emoji is not None):
            raise ValueError(
                f'`icon` and `unicode_emoji` parameters are mutually exclusive, got '
                f'icon = {icon!r}; unicode_emoji = {unicode_emoji!r}.'
            )
        
        
        self = object.__new__(cls)
        self.color = color
        self.color_configuration = color_configuration
        self.flags = flags
        self.guild_id = 0
        self.icon = icon
        self.id = 0
        self.manager_metadata = manager_metadata
        self.manager_type = manager_type
        self.mentionable = mentionable
        self.name = name
        self.permissions = permissions
        self.position = position
        self.separated = separated
        self.unicode_emoji = unicode_emoji
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0, *, strong_cache = True):
        """
        Creates a role from the given `data` at the given `guild`. If the role already exists and is not partial, then
        returns it. However it is partial, then updates it as well.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Role data received from Discord.
        
        guild_id : `int` = `0`, Optional
            The owner guild's identifier.
        
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        self : `instance<cls>`
        """
        role_id = parse_id(data)
        
        try:
            self = ROLES[role_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = role_id
            self.guild_id = guild_id
            self._set_attributes(data)
            ROLES[role_id] = self
        
        else:
            if strong_cache and (not self.partial):
                return self
        
            self.guild_id = guild_id
            self._set_attributes(data)
        
        if strong_cache:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                guild.roles[role_id] = self
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the role into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        # color
        put_color(self.color, data, defaults)
        
        # color_configuration
        put_color_configuration(self.color_configuration, data, defaults)
        
        # flags
        put_flags(self.flags, data, defaults)
        
        # icon
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        
        # id
        if include_internals:
            put_id(self.id, data, defaults)
        
        # manager
        if include_internals:
            put_manager((self.manager_type, self.manager_metadata), data, defaults)
        
        # mentionable
        put_mentionable(self.mentionable, data, defaults)
        
        # name
        put_name(self.name, data, defaults)
        
        # permissions
        put_permissions(self.permissions, data, defaults)
        
        # position
        put_position(self.position, data, defaults)
        
        # separated
        put_separated(self.separated, data, defaults)
        
        # unicode_emoji
        put_unicode_emoji(self.unicode_emoji, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the role's attributes from the given data.
        
        `.id` and `.guild_id` fields should be already set,
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received role data.
        """
        self.color = parse_color(data)
        self.color_configuration = parse_color_configuration(data)
        self.flags = parse_flags(data)
        self._set_icon(data)
        self.manager_type, self.manager_metadata = parse_manager(data)
        self.mentionable = parse_mentionable(data)
        self.name = parse_name(data)
        self.permissions = parse_permissions(data)
        self.position = parse_position(data)
        self.separated = parse_separated(data)
        self.unicode_emoji = parse_unicode_emoji(data)

    
    @classmethod
    def precreate(cls, role_id, *, guild_id = ..., **keyword_parameters):
        """
        Precreates a role by creating a partial one with the given parameters. When the role is loaded, then the
        precreated one will be picked up. However if the role is precreated when it already exists, then the existing
        one is picked up and is updated by the given parameters only if it is partial.
        
        Parameters
        ----------
        role_id : `int`, `str`
            The role's id.
        
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The role's guild or its identifier.
        
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the role.
        
        Other Parameters
        ----------------
        color : ``None | int | Color``, Optional (Keyword only)
            The role's color.
        
        color_configuration : ``None | RoleColorConfiguration``, Optional (Keyword only)
            The role's color's configuration.
        
        flags : ``None | int | RoleFlag``, Optional (Keyword only)
            The role's flags.
        
        icon : ``None | str | Icon``, Optional (Keyword only)
            The role's icon.
            
            > Mutually exclusive with the `unicode_emoji` parameter.
        
        manager : ``None | (RoleManagerType, RoleManagerMetadataBase)``, Optional (Keyword only)
            The role's manager.
        
        manager_metadata : `None`, ``RoleManagerMetadataBase``, Optional (Keyword only)
            The role's manager's metadata.
        
        manager_type : ``RoleManagerType``, Optional (Keyword only)
            TIf the role is managed, defines by what it is.
        
        mentionable : `bool`, Optional (Keyword only)
            Whether the role can be mentioned.
        
        name : `str`, Optional (Keyword only)
            The name of the role.
        
        separated : `bool`, Optional (Keyword only)
            Users show up in separated groups by their highest `separated` role.
        
        permissions : `int`, ``Permission``, Optional (Keyword only)
            The permissions of the users having the role.
        
        position : `int`, Optional (Keyword only)
            The permissions of the users having the role.
        
        unicode_emoji : ``None | Emoji``
            The role's icon as an unicode emoji.
            
            > Mutually exclusive with the `icon`, `icon_type` and `icon_hash` parameters.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        role_id = validate_id(role_id)
        
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        if keyword_parameters:
            processed = []
            
            # icon
            icon = cls.icon.parse_from_keyword_parameters(keyword_parameters)
            if (icon is not None):
                processed.append(('icon', icon))
            
            # unicode_emoji
            try:
                unicode_emoji = keyword_parameters.pop('unicode_emoji')
            except KeyError:
                unicode_emoji = None
            else:
                unicode_emoji = validate_unicode_emoji(unicode_emoji)
                if (unicode_emoji is not None):
                    processed.append(('unicode_emoji', unicode_emoji))
            
            # Mutually exclusive check
            if (icon is not None) and (unicode_emoji is not None):
                raise ValueError(
                    f'`icon` and `unicode_emoji` parameters are mutually exclusive, got '
                    f'icon = {icon!r}; unicode_emoji = {unicode_emoji!r}.'
                )
            
            # manager
            try:
                manager = keyword_parameters.pop('manager')
            except KeyError:
                pass
            else:
                manager_type, manager_metadata = validate_manager(manager)
                processed.append(('manager_type', manager_type))
                processed.append(('manager_metadata', manager_metadata))
            
            extra = process_precreate_parameters(keyword_parameters, PRECREATE_FIELDS, processed)
            raise_extra(extra)
        
        else:
            processed = None
        
        try:
            self = ROLES[role_id]
        except KeyError:
            self = cls._create_empty(role_id, guild_id)
            ROLES[role_id] = self
        else:
            if (not self.partial):
                return self
            
            if guild_id and (not self.guild_id):
                self.guild_id = guild_id
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def _update_attributes(self, data):
        """
        Updates the role with the given `data` with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Role data received from Discord.
        """
        clear_cache_permission = False
        
        # color
        self.color = parse_color(data)
        
        # color_configuration
        self.color_configuration = parse_color_configuration(data)
        
        # flags
        self.flags = parse_flags(data)
        
        # icon
        self._set_icon(data)
        
        # manager
        if self.manager_type is RoleManagerType.unset:
            self.manager_type, self.manager_metadata = parse_manager(data)
        
        # mentionable
        self.mentionable = parse_mentionable(data)
        
        # name
        self.name = parse_name(data)
        
        # permissions
        permissions = Permission(data[PERMISSION_KEY])
        if self.permissions != permissions:
            self.permissions = permissions
            clear_cache_permission = True
        
        # position
        position = parse_position(data)
        if self.position != position:
            self.position = position
            clear_cache_permission = True
        
        # separated
        self.separated = parse_separated(data)
        
        # unicode_emoji
        self.unicode_emoji = parse_unicode_emoji(data)
        
        if clear_cache_permission:
            self._clear_cache_permission()
    
    
    def _clear_cache_permission(self):
        """
        Clears the role's guild's permission cache.
        """
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                guild._invalidate_cache_permission()
    
    
    def __repr__(self):
        """Returns the role's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        role_id = self.id
        if role_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(role_id))
            repr_parts.append(',')
            
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the role with the given `data` and returns it's overwritten attributes as a `dict` with a
        `attribute-name` - `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Role data received from Discord.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-----------------------+---------------------------+
        | Keys                  | Values                    |
        +=======================+===========================+
        | color                 | ``Color``                 |
        +-----------------------+---------------------------+
        | color_configuration   | ``ColorConfiguration``    |
        +-----------------------+---------------------------+
        | flags                 | ``RoleFlag``              |
        +-----------------------+---------------------------+
        | icon                  | ``Icon``                  |
        +-----------------------+---------------------------+
        | mentionable           | `bool`                    |
        +-----------------------+---------------------------+
        | name                  | `str`                     |
        +-----------------------+---------------------------+
        | permissions           | ``Permission``            |
        +-----------------------+---------------------------+
        | position              | `int`                     |
        +-----------------------+---------------------------+
        | separated             | `bool`                    |
        +-----------------------+---------------------------+
        | unicode_emoji         | ``None | Emoji``          |
        +-----------------------+---------------------------+
        """
        old_attributes = {}
        
        clear_cache_permission = False
        
        # color
        color = parse_color(data)
        if self.color != color:
            old_attributes['color'] = self.color
            self.color = color
        
        # color_configuration
        color_configuration = parse_color_configuration(data)
        if self.color_configuration != color_configuration:
            old_attributes['color_configuration'] = self.color_configuration
            self.color_configuration = color_configuration
        
        # flags
        flags = parse_flags(data)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        # icon
        self._update_icon(data, old_attributes)
        
        # managed
        if self.manager_type is RoleManagerType.unset:
            self.manager_type, self.manager_metadata = parse_manager(data)
        
        # mentionable
        mentionable = parse_mentionable(data)
        if self.mentionable != mentionable:
            old_attributes['mentionable'] = self.mentionable
            self.mentionable = mentionable
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # permissions
        permissions = parse_permissions(data)
        if self.permissions != permissions:
            old_attributes['permissions'] = self.permissions
            self.permissions = permissions
            clear_cache_permission = True
        
        # position
        position = parse_position(data)
        if self.position != position:
            old_attributes['position'] = self.position
            self.position = position
            clear_cache_permission = True
        
        # separated
        separated = parse_separated(data)
        if self.separated != separated:
            old_attributes['separated'] = self.separated
            self.separated = separated
        
        # unicode_emoji
        unicode_emoji = parse_unicode_emoji(data)
        if self.unicode_emoji is not unicode_emoji:
            old_attributes['unicode_emoji'] = self.unicode_emoji
            self.unicode_emoji = unicode_emoji
        
        if clear_cache_permission:
            self._clear_cache_permission()
        
        return old_attributes
    
    
    def _delete(self):
        """
        Removes the role's references.
        
        Called when the role is deleted.
        """
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            guild._invalidate_cache_permission()
            
            role_id = self.id
            try:
                del guild.roles[role_id]
            except KeyError:
                pass
            else:
                for user in guild.users.values():
                    try:
                        guild_profile = user.guild_profiles[guild_id]
                    except KeyError:
                        pass
                    else:
                        role_ids = guild_profile.role_ids
                        if (role_ids is not None):
                            try:
                                index = role_ids.index(role_id)
                            except ValueError:
                                pass
                            else:
                                if len(role_ids) == 1:
                                    role_ids = None # It is so deep! (that's what she said)
                                else:
                                    role_ids = (*role_ids[:index], *role_ids[index + 1:])
                                guild_profile.role_ids = role_ids
    
    
    def is_default(self):
        """
        Returns whether the role is the default role at it's respective guild.
        
        Returns
        -------
        is_default : `bool`
        """
        return (self.id == self.guild_id)
    
    
    @property
    def mention(self):
        """
        The role's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<@&{self.id}>'
    
    
    def __format__(self, code):
        """
        Formats the role in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        representation : `str`
        
        Examples
        --------
        ```py
        >>>> from hata import Role, now_as_id
        >>>> role = Role.precreate(now_as_id(), name = 'admiral-general')
        >>>> role
        <Role name = 'admiral-general', id = 725333995067277312>
        >>>> # no code stands for `role.name`.
        >>>> f'{role}'
        'admiral-general'
        >>>> # 'm' stands for mention.
        >>>> f'{role:m}'
        '<@&725333995067277312>'
        >>>> # 'c' stands for created at.
        >>>> f'{role:c}'
        '2020.06.24-12:58:20'
        ```
        """
        if not code:
            return self.name
        
        if code == 'm':
            return self.mention
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {type(self).__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}, {"m"!r}.'
        )
    
    
    @property
    def users(self):
        """
        Returns the users who have the role.
        
        Returns
        -------
        users : ``list<ClientUserBase>``
        """
        users = []
        guild_id = self.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            role_id = self.id
            if role_id == guild.id:
                users.extend(guild.users.values())
            else:
                for user in guild.users.values():
                    try:
                        guild_profile = user.guild_profiles[guild.id]
                    except KeyError:
                        # should not happen
                        pass
                    else:
                        role_ids = guild_profile.role_ids
                        if (role_ids is not None):
                            if role_id in role_ids:
                                users.append(user)
        
        return users
    
    
    @property
    def managed(self):
        """
        Returns whether the role is managed.
        
        Returns
        -------
        managed: `bool`
        """
        return (self.manager_type is not RoleManagerType.none)
    
    
    @property
    @copy_docs(RoleManagerMetadataBase.manager)
    def manager(self):
        return self.manager_metadata.manager
    
    
    @property
    @copy_docs(RoleManagerMetadataBase.manager_id)
    def manager_id(self):
        return self.manager_metadata.manager_id
    
    
    @property
    def guild(self):
        """
        Returns the role's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def partial(self):
        """
        Returns whether the role is partial.
        
        Partial roles have `.guild` set as ``None``.
        
        Returns
        -------
        partial : `bool`
        """
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                if (not guild.partial) and (self.id in guild.roles):
                    return False
        
        return True
    
    
    def __eq__(self, other):
        """Returns whether the two roles are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id == other_id
        
        return self._is_equal_partial(other)
    
    
    def __ne__(self, other):
        """Returns whether the two roles are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id != other_id
        
        return not self._is_equal_partial(other)
    
    
    def __gt__(self, other):
        """Returns whether this role's position is higher than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.position > other.position:
            return True
        
        if self.position == other.position:
            if self.id > other.id:
                return True
    
        return False
    
    
    def __ge__(self, other):
        """Returns whether this role's position is higher or equal to the other's."""
        if type(self) is not type(other):
            return NotImplemented
            
        if self.position > other.position:
            return True
        
        if self.position == other.position:
            if self.id >= other.id:
                return True
        
        return False
    
        
    def __le__(self, other):
        """Returns whether this role's position is less or equal to the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.position < other.position:
            return True
        
        if self.position == other.position:
            if self.id <= other.id:
                return True
        
        return False
        
    
    
    def __lt__(self, other):
        """Returns whether this role's position is less than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.position < other.position:
            return True
        
        if self.position == other.position:
            if self.id < other.id:
                return True
        
        return False
    
    
    def __hash__(self):
        """Returns the role's hash value."""
        role_id = self.id
        if role_id:
            return role_id
        
        return self._get_hash_partial()
    
    
    @classmethod
    def _create_empty(cls, role_id, guild_id):
        """
        Creates an empty role with the given identifier.
        
        Parameters
        ----------
        role_id : `int`
            The role's identifier.
        guild_id : `int`
            The role's guild's identifier.
        
        Returns
        -------
        self : ``Role``
            The created role.
        """
        self = object.__new__(cls)
        self.id = role_id
        
        self.color = Color()
        self.color_configuration = RoleColorConfiguration.create_empty()
        self.flags = RoleFlag()
        self.guild_id = guild_id
        self.separated = False
        self.manager_metadata = ROLE_MANAGER_DEFAULT
        self.manager_type = RoleManagerType.none
        self.mentionable = False
        self.name = ''
        self.permissions = PERMISSION_NONE
        self.position = 1
        self.icon_type = ICON_TYPE_NONE
        self.icon_hash = 0
        self.unicode_emoji = None
        
        return self
    
    
    def _is_equal_partial(self, other):
        """
        Returns whether the role is equal to the given one.
        This function is called when one or both the roles are partial.
        
        Parameters
        ----------
        other : ``instance<type<self>>``
            The other role to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # color
        if self.color != other.color:
            return False
        
        # color_configuration
        if self.color_configuration != other.color_configuration:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # icon
        if self.icon != other.icon:
            return False
        
        # guild_id
        # Skip it, non partial field
        
        # id
        # Skip it, non partial field
        
        # manager_metadata
        if self.manager_metadata != other.manager_metadata:
            return False
        
        # manager_type
        if self.manager_type != other.manager_type:
            return False
        
        # mentionable
        if self.mentionable != other.mentionable:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # permissions
        if self.permissions != other.permissions:
            return False
        
        # position
        if self.position != other.position:
            return False
        
        # separated
        if self.separated != other.separated:
            return False
        
        # unicode_emoji
        if self.unicode_emoji is not other.unicode_emoji:
            return False
        
        return True
    
    
    def _get_hash_partial(self):
        """
        Returns the role's hash value.
        This function is called when the role is partial.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # color
        hash_value ^= self.color
        
        # color_configuration
        hash_value ^= hash(self.color_configuration)
        
        # flags
        hash_value ^= self.flags << 16
        
        # icon
        hash_value ^= hash(self.icon)
        
        # guild_id
        # Skip it, non partial field
        
        # id
        # Skip it, non partial field
        
        # manager_metadata
        hash_value ^= hash(self.manager_metadata)
        
        # manager_type
        hash_value ^= hash(self.manager_type) << 24
        
        # mentionable
        hash_value ^= self.mentionable << 28
        
        # name
        hash_value ^= hash(self.name)
        
        # permissions
        hash_value ^= self.permissions
        
        # position
        hash_value ^= self.position << 29
        
        # separated
        hash_value ^= self.separated << 5
        
        # unicode_emoji
        unicode_emoji = self.unicode_emoji
        if (unicode_emoji is not None):
            hash_value ^= self.unicode_emoji.id << 6
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the role returning a new partial one.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        new.color = self.color
        new.color_configuration = self.color_configuration.copy()
        new.flags = self.flags
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        new.guild_id = 0
        new.id = 0
        new.manager_metadata = self.manager_metadata.copy()
        new.manager_type = self.manager_type
        new.mentionable = self.mentionable
        new.name = self.name
        new.permissions = self.permissions
        new.position = self.position
        new.separated = self.separated
        new.unicode_emoji = self.unicode_emoji
        return new
    
    
    def copy_with(
        self,
        *,
        color = ...,
        color_configuration = ...,
        flags = ...,
        icon = ...,
        manager = ...,
        mentionable = ...,
        name = ...,
        permissions = ...,
        position = ...,
        separated = ...,
        unicode_emoji = ...,
    ):
        """
        Copies the role with modifying its defined fields, returning a new partial one.
        
        Parameters
        ----------
        color : ``None | int | Color``, Optional (Keyword only)
            The role's color.
        
        color_configuration : ``None | RoleColorConfiguration``, Optional (Keyword only)
            The role's color's configuration.
        
        flags : ``None | int | RoleFlag``, Optional (Keyword only)
            The role's flags.
        
        icon : ``None | str | bytes-like | Icon``, Optional (Keyword only)
            The role's icon.
            
            > Mutually exclusive with the `unicode_emoji` parameter.
        
        manager : ``None | (RoleManagerType, RoleManagerMetadataBase)``, Optional (Keyword only)
            The role's manager.
        
        mentionable : `bool`, Optional (Keyword only)
            Whether the role can be mentioned.
        
        name : `str`, Optional (Keyword only)
            The role's name.
        
        permissions : `int`, ``Permission``, Optional (Keyword only)
            The permissions of the users having the role.
        
        position : `int`, Optional (Keyword only)
            The role's position.
        
        separated : `bool`, Optional (Keyword only)
            Users show up in separated groups by their highest `separated` role.
        
        unicode_emoji : ``None | Emoji``, Optional (Keyword only)
            The role's icon as an unicode emoji.
            
            > Mutually exclusive with the `icon` parameter.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Value of invalid type given.
        ValueError
            - Value of invalid value given.
            - `icon` and `unicode_emoji` are mutually exclusive.
        """
        # Make sure by passing only either `icon` / `unicode_emoji` we wont raise error by nulling out the existing.
        if (icon is ...):
            self_unicode_emoji = self.unicode_emoji
        else:
            self_unicode_emoji = None
        
        if (unicode_emoji is ...):
            self_icon = self.icon
        else:
            self_icon = None
        
        # color
        if color is ...:
            color = self.color
        else:
            color = validate_color(color)
        
        # color_configuration
        if color_configuration is ...:
            color_configuration = self.color_configuration.copy()
        else:
            color_configuration = validate_color_configuration(color_configuration)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # icon
        if icon is ...:
            icon = self_icon
        else:
            icon = type(self).icon.validate_icon(icon, allow_data = True)
        
        # manager
        if manager is ...:
            manager_type = self.manager_type
            manager_metadata = self.manager_metadata.copy()
        else:
            manager_type, manager_metadata = validate_manager(manager)
        
        # mentionable
        if mentionable is ...:
            mentionable = self.mentionable
        else:
            mentionable = validate_mentionable(mentionable)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # permissions
        if permissions is ...:
            permissions = self.permissions
        else:
            permissions = validate_permissions(permissions)
        
        # position
        if position is ...:
            position = self.position
        else:
            position = validate_position(position)
        
        # separated
        if separated is ...:
            separated = self.separated
        else:
            separated = validate_separated(separated)
        
        # unicode_emoji
        if unicode_emoji is ...:
            unicode_emoji = self_unicode_emoji
        else:
            unicode_emoji = validate_unicode_emoji(unicode_emoji)
        
        if (icon is not None) and (unicode_emoji is not None):
            raise ValueError(
                f'`icon` and `unicode_emoji` parameters are mutually exclusive, got '
                f'icon = {icon!r}; unicode_emoji = {unicode_emoji!r}.'
            )
        
        
        new = object.__new__(type(self))
        new.color = color
        new.color_configuration = color_configuration
        new.flags = flags
        new.guild_id = 0
        new.icon = icon
        new.id = 0
        new.manager_metadata = manager_metadata
        new.manager_type = manager_type
        new.mentionable = mentionable
        new.name = name
        new.permissions = permissions
        new.position = position
        new.separated = separated
        new.unicode_emoji = unicode_emoji
        return new
    
    
    @property
    def icon_url(self):
        """
        Returns the role's icon's url. If the role has no icon, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_role_icon_url(self.id, self.icon_type, self.icon_hash)
    
    
    def icon_url_as(self, ext = None, size = None):
        """
        Returns the role's icon's url. If the role has no icon, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_role_icon_url_as(self.id, self.icon_type, self.icon_hash, ext, size)



__all__ = ()

from ...bases import Icon
from ...color import Color
from ...emoji import create_unicode_emoji
from ...permission import Permission
from ...permission.constants import PERMISSION_KEY
from ...role import RoleColorConfiguration, RoleFlag
from ...role.role.fields import (
    validate_color, validate_color_configuration, validate_flags, validate_mentionable, validate_name,
    validate_permissions, validate_position, validate_separated, validate_unicode_emoji
)
from ...role.role.role import ROLE_ICON

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_name, value_serializer_name


# ---- color ----

COLOR_CONVERSION = AuditLogEntryChangeConversion(
    ('color',),
    'color',
    value_validator = validate_color,
)


@COLOR_CONVERSION.set_value_deserializer
def color_value_deserializer(value):
    if value is None:
        value = Color()
    else:
        value = Color(value)
    
    return value


@COLOR_CONVERSION.set_value_serializer
def color_value_serializer(value):
    return int(value)


# ---- color_configuration ----

COLOR_CONFIGURATION_CONVERSION = AuditLogEntryChangeConversion(
    ('colors',),
    'color_configuration',
    value_validator = validate_color_configuration,
)


@COLOR_CONFIGURATION_CONVERSION.set_value_deserializer
def color_configuration_value_deserializer(value):
    if value is None:
        value = RoleColorConfiguration.create_empty()
    else:
        value = RoleColorConfiguration.from_data(value)
    
    return value


@COLOR_CONFIGURATION_CONVERSION.set_value_serializer
def color_configuration_value_serializer(value):
    return value.to_data(defaults = True)


# ---- flags ----

FLAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('flags',),
    'flags',
    value_validator = validate_flags,
)


@FLAGS_CONVERSION.set_value_deserializer
def flags_value_deserializer(value):
    if value is None:
        value = RoleFlag()
    else:
        value = RoleFlag(value)
    
    return value


@FLAGS_CONVERSION.set_value_serializer
def flags_value_serializer(value):
    return int(value)


# ---- icon ----

ICON_CONVERSION = AuditLogEntryChangeConversion(
    ('icon_hash',),
    'icon',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = ROLE_ICON.validate_icon,
)


# ---- mentionable ----

MENTIONABLE_CONVERSION = AuditLogEntryChangeConversion(
    ('mentionable',),
    'mentionable',
    value_validator = validate_mentionable,
)


@MENTIONABLE_CONVERSION.set_value_deserializer
def mentionable_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- permissions ----

PERMISSIONS_CONVERSION = AuditLogEntryChangeConversion(
    (PERMISSION_KEY,),
    'permissions',
    value_validator = validate_permissions,
)


@PERMISSIONS_CONVERSION.set_value_deserializer
def permission_value_deserializer(value):
    if value is None:
        value = Permission()
    else:
        value = Permission(value)
    return value


@PERMISSIONS_CONVERSION.set_value_serializer
def permission_value_serializer(value):
    return format(value, 'd')


# ---- position ----

POSITION_CONVERSION = AuditLogEntryChangeConversion(
    ('position',),
    'position',
    value_validator = validate_position,
)

@POSITION_CONVERSION.set_value_deserializer
def position_value_deserializer(value):
    if value is None:
        value = 0
    return value


# ---- separated ----

SEPARATED_CONVERSION = AuditLogEntryChangeConversion(
    ('hoist',),
    'separated',
    value_validator = validate_separated,
)


@SEPARATED_CONVERSION.set_value_deserializer
def separated_value_deserializer(value):
    if value is None:
        value = False

    return value


# ---- unicode_emoji ----

UNICODE_EMOJI_CONVERSION = AuditLogEntryChangeConversion(
    ('unicode_emoji',),
    'unicode_emoji',
    value_validator = validate_unicode_emoji,
)

@UNICODE_EMOJI_CONVERSION.set_value_deserializer
def position_value_deserializer(value):
    if (value is not None):
        value = create_unicode_emoji(value)
    
    return value


@UNICODE_EMOJI_CONVERSION.set_value_serializer
def position_value_serializer(value):
    if (value is not None):
        value = value.unicode
    
    return value


# ---- Construct ----

ROLE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    COLOR_CONVERSION,
    COLOR_CONFIGURATION_CONVERSION,
    FLAGS_CONVERSION,
    ICON_CONVERSION,
    MENTIONABLE_CONVERSION,
    NAME_CONVERSION,
    PERMISSIONS_CONVERSION,
    POSITION_CONVERSION,
    SEPARATED_CONVERSION,
    UNICODE_EMOJI_CONVERSION,
)



__all__ = ('SlashCommandParameterConversionError', 'SlasherCommandError', 'SlasherSyncError')

from random import choice

from scarletio import CallableAnalyzer, RichAttributeErrorBaseType, include
from scarletio import copy_docs

from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.interaction import InteractionType

from .responding import InteractionResponse

SlashCommand = include('SlashCommand')
Slasher = include('Slasher')


class SlasherCommandError(Exception, RichAttributeErrorBaseType):
    """
    Base class for slash command internal errors.
    """
    __slots__ = ()
    __init__ = object.__init__
    
    
    def __new__(cls):
        return Exception.__new__(cls)
    
    
    def __repr__(self):
        """Returns the exception's representation."""
        return ''.join(['<', type(self).__name__, '>'])
    
    
    @property
    def pretty_repr(self):
        """
        Returns the pretty representation of the exception.
        
        Returns
        -------
        representation : `str`
        """
        return ''
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return True

    
    def __hash__(self):
        """Returns the exception's hash value."""
        return 0


class SlashCommandParameterConversionError(SlasherCommandError):
    """
    Exception raised when a command's parameter's parsing fails.
    
    Attributes
    ----------
    _pretty_repr : `None`, `str`
        generated pretty representation of the exception.
    parameter_name : `None`, `str`
        The parameter's name, which failed to be parsed.
    received_value : `None`, `str`
        The parameter's received value.
    excepted_type : `None`, `str`
        The parameter's expected type's name.
    expected_values : `None`, `list` of `object`
        Expected values.
    """
    __slots__ = ('_pretty_repr', 'excepted_type', 'expected_values', 'parameter_name', 'received_value')
    
    def __new__(cls, parameter_name, received_value, excepted_type, expected_values):
        """
        Creates a new parameter conversion error with the given parameters.
        
        Parameters
        ----------
        parameter_name : `None`, `str`
            The parameter's name, which failed to be parsed.
        received_value : `None`, `str`
            The parameter's received value.
        excepted_type : `None`, `str`
            The parameter's expected type's name.
        expected_values : `None`, `list` of `object`
            Expected values.
        """
        self = Exception.__new__(cls, parameter_name, received_value, excepted_type, expected_values)
        self.parameter_name = parameter_name
        self.received_value = received_value
        self.excepted_type = excepted_type
        self.expected_values = expected_values
        self._pretty_repr = None
        return self
    
    
    @copy_docs(SlasherCommandError.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # parameter_name
        parameter_name = self.parameter_name
        if (parameter_name is not None):
            field_added = True
            
            repr_parts.append(' parameter_name = ')
            repr_parts.append(repr(parameter_name))
        
        # excepted_type
        excepted_type = self.excepted_type
        if (excepted_type is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' excepted_type = ')
            repr_parts.append(repr(excepted_type))
        
        # expected_values
        expected_values = self.expected_values
        if (expected_values is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' expected_values = ')
            repr_parts.append(repr(expected_values))
        
        # received_value
        if field_added:
            repr_parts.append(',')
        else:
            field_added = True
        
        repr_parts.append(' received_value = ')
        repr_parts.append(repr(self.received_value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    @copy_docs(SlasherCommandError.pretty_repr)
    def pretty_repr(self):
        pretty_repr = self._pretty_repr
        if pretty_repr is None:
            pretty_repr = self._create_pretty_repr()
            self._pretty_repr = pretty_repr
        
        return pretty_repr
    
    
    def _create_pretty_repr(self):
        """
        Creates the pretty representation of the parameter conversion error.
        
        Returns
        -------
        repr_ : `str`
            The representation of the parameter conversion error.
        """
        repr_parts = ['Parameter conversion failed\n']
        
        parameter_name = self.parameter_name
        if (parameter_name is not None):
            repr_parts.append('\n')
            repr_parts.append('Name: `')
            repr_parts.append(parameter_name)
            repr_parts.append('`')
        
        excepted_type = self.excepted_type
        if (excepted_type is not None):
            repr_parts.append(
                '\n'
                'Excepted type: `'
            )
            repr_parts.append(excepted_type)
            repr_parts.append('`')
    
        
        expected_values = self.expected_values
        if (expected_values is not None):
            repr_parts.append(
                '\n'
                'Expected value(s):'
            )
            
            for expected_value in expected_values:
                repr_parts.append('\n- `')
                repr_parts.append(str(expected_value))
                repr_parts.append('`')
        
        repr_parts.append(
            '\n'
            'Received: '
        )
        received_value = self.received_value
        if (received_value is None):
            repr_parts.append('N/A')
        else:
            repr_parts.append('`')
            repr_parts.append(repr(received_value))
            repr_parts.append('`')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(SlasherCommandError.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # parameter_name
        if self.parameter_name != other.parameter_name:
            return False
        
        # received_value
        if self.received_value != other.received_value:
            return False
        
        # expected_type
        if self.excepted_type != other.excepted_type:
            return False
        
        # expected_values
        if self.expected_values != other.expected_values:
            return False
        
        return True
    
    
    @copy_docs(SlasherCommandError.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # parameter_name
        parameter_name = self.parameter_name
        if (parameter_name is not None):
            hash_value ^= hash(parameter_name)
        
        # received_value
        received_value = self.received_value
        if (received_value is not None):
            hash_value ^= hash(received_value)
        
        # excepted_type
        excepted_type = self.excepted_type
        if (excepted_type is not None):
            hash_value ^= hash(excepted_type)
        
        # expected_values
        expected_values = self.expected_values
        if (expected_values is not None):
            hash_value ^= len(expected_values)
            
            for expected_value in expected_values:
                hash_value ^= hash(expected_value)
        
        return hash_value


ERROR_MESSAGES = [
    (
       'Exception occurred meanwhile processing your interaction.\n'
       'Our highly educated Cirno-s are already working on the problem.'
    ), (
        'Your command\'s execution failed.\n'
        'We are putting Reimu to investigate about the issue.'
    ), (
        'The command died out.\n'
        'Maybe Izaoyi love-shop has some antidote.'
    ), (
        'Something went wrong.\n'
        'Do you need driver license for broom riding?'
    ),
]

def default_slasher_random_error_message_getter():
    """
    Returns random error message getter of ``Slasher``.
    
    Returns
    -------
    error_message : `str`
    """
    return choice(ERROR_MESSAGES)


def _validate_random_error_message_getter(random_error_message_getter):
    """
    Validates the given `random_error_message_getter`. It should accept no parameters and return a string.
    
    Parameters
    ----------
    random_error_message_getter : `callable`
        The random message getter to validate.
    
    Raises
    ------
    TypeError
        - If `random_error_message_getter` is not callable.
        - If `random_error_message_getter` is not async.
        - If `random_error_message_getter` excepts not `0` parameters.
    """
    analyzer = CallableAnalyzer(random_error_message_getter)
    if analyzer.is_async():
        raise TypeError(
            f'`random_error_message_getter` cannot be async, got {random_error_message_getter!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if (min_ > 0):
        raise TypeError(
            f'A `random_error_message_getter` should accept `0` parameters, meanwhile '
            f'{random_error_message_getter!r} accepts between `{min_}` and `{max_}`.'
        )


async def default_slasher_exception_handler(client, interaction_event, command, exception):
    """
    Default ``Slasher`` exception handler.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    interaction_event : ``InteractionEvent``
        The received interaction event.
    command : ``SlashCommand``, ``ComponentCommand``
        The command, which raised.
    exception : `BaseException`
        The occurred exception.
    
    Returns
    -------
    handled : `bool`
        Whether the error handler handled the exception.
    """
    render = True
    
    try:
        if isinstance(exception, SlasherCommandError):
            forward = exception.pretty_repr
            render = False
            create_new_message = False
        
        elif isinstance(exception, DiscordException) and (exception.status == 500):
            forward = None
            create_new_message = False
        
        elif (interaction_event.type is InteractionType.application_command) and interaction_event.is_unanswered():
            forward = client.slasher._random_error_message_getter()
            create_new_message = False
        
        elif (interaction_event.type is InteractionType.message_component) and interaction_event.is_unanswered():
            forward = client.slasher._random_error_message_getter()
            create_new_message = True
            
            try:
                await client.interaction_component_acknowledge(interaction_event)
            except GeneratorExit:
                raise
            
            except ConnectionError:
                pass
            
            except DiscordException as err:
                if not (
                    (err.status == 500) or
                    (
                        err.code in (
                            ERROR_CODES.unknown_interaction,
                            ERROR_CODES.unknown_webhook,
                        )
                    )
                ):
                    raise
        
        else:
            forward = None
            create_new_message = False
        
        if (forward is not None):
            # Process what to forward
            if isinstance(forward, tuple):
                forward = InteractionResponse(*forward)
            else:
                forward = InteractionResponse(forward)
            
            # Get forward function
            if create_new_message:
                function = type(client).interaction_followup_message_create
            else:
                function = type(client).interaction_response_message_create
            
            try:
                await function(
                    client,
                    interaction_event,
                    forward,
                    show_for_invoking_user_only = True,
                )
            except GeneratorExit:
                raise
            
            except ConnectionError:
                pass
            
            except DiscordException as err:
                if not (
                    (err.status == 500) or
                    (
                        err.code in (
                            ERROR_CODES.unknown_interaction,
                            ERROR_CODES.unknown_webhook,
                        )
                    )
                ):
                    raise
    
    except GeneratorExit:
        # Disable `await` in finally.
        render = False
        raise
    
    finally:
        if render:
            await _render_application_command_exception(client, command, exception)
    
    return True


async def _render_application_command_exception(client, command, exception):
    """
    Renders interaction command exception.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    command : ``SlashCommand``, ``ComponentCommand``
        The command, which raised.
    exception : `BaseException`
        The occurred exception.
    """
    await client.events.error(client, f'`Slasher` while calling `{command!r}`', exception)


async def handle_command_exception(entity, client, interaction_event, exception):
    """
    handles slash command exception.
    
    This function is a coroutine.

    Parameters
    ----------
    entity : ``ComponentCommand``, ``Slasher``, ``EmbeddedActivityLaunchCommand``, ``SlashCommand``,
            ``SlashCommandCategory``, ``SlashCommandFunction``, ``SlashCommandParameterAutoCompleter``
        The entity to iterate it's exception handlers.
    client : ``Client``
        The respective client.
    interaction_event : ``InteractionEvent``
        The received interaction event.
    exception : `BaseException`
        The occurred exception.
    """
    for exception_handler in _iter_exception_handlers(entity):
        try:
            handled = await exception_handler(client, interaction_event, entity, exception)
        except GeneratorExit:
            raise
        
        except BaseException as err:
            await client.events.error(client, 'handle_command_exception', err)
            return
        
        if handled:
            return
    
    if not isinstance(exception, SlasherCommandError):
        await _render_application_command_exception(client, entity, exception)


def _iter_exception_handlers(entity):
    """
    Iterates the given entity's exception handlers.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    entity : ``ComponentCommand``, ``Slasher``, ``EmbeddedActivityLaunchCommand``, ``SlashCommand``,
            ``SlashCommandCategory``, ``SlashCommandFunction``, ``SlashCommandParameterAutoCompleter``
        The entity to iterate it's exception handlers.
    
    Yields
    ------
    exception_handler : `CoroutineFunction`
    """
    while True:
        exception_handlers = entity._exception_handlers
        if (exception_handlers is not None):
            yield from exception_handlers
        
        if isinstance(entity, Slasher):
            break
        
        parent_reference = entity._parent_reference
        if parent_reference is None:
            break
        
        entity = parent_reference()
        if entity is None:
            break
    


def _produce_slash_sync_error_representation(exception):
    """
    Helper function to produce slash sync error representation.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    exception : ``SlasherSyncError``
        Exception to produce its representation of.
    
    Yields
    ------
    part : `str`
    """
    yield type(exception).__name__
    
    entity = exception.entity
    if (entity is not None):
        yield ' while syncing '
        yield entity.name
        
        yield '\nentity = '
        yield repr(entity)
    
    cause = exception.__cause__
    if (cause is not None):
        yield '\ncause = '
        yield type(cause).__name__
        
        if isinstance(cause, DiscordException):
            if (
                (cause.status == 400) and
                (cause.code == ERROR_CODES.invalid_form_body) and
                (cause.message == 'redirect_uris[0].BASE_TYPE_REQUIRED(\'This field is required\')')
            ):
                yield (
                    '\n'
                    '! This error is due to the application having a null value in its redirect urls !\n'
                    '! You have to go to Discord Developer Portal and force update the field by changing it !'
                )


class SlasherSyncError(BaseException):
    """
    Raised when syncing an application command or it's permission fails.
    
    The main purpose of the exception is to include the entity in context within the traceback.
    
    Attributes
    ----------
    entity: ``None | SlashCommand``
        The entity, who's sync failed.
    """
    __slots__ = ('entity',)
    
    def __new__(cls, entity, cause):
        """
        Creates a new slasher sync error exception.
        
        Parameters
        ----------
        entity: ``None | SlashCommand``
            The entity, who's sync failed.
        
        cause : `None |BaseException`
            Source exception.
        """
        self = BaseException.__new__(cls, entity)
        self.entity = entity
        self.__cause__ = cause
        return self
    
    
    __init__ = object.__init__
    
    
    def __repr__(self):
        """Returns repr(self)."""
        return ''.join(_produce_slash_sync_error_representation(self))
    
    
    __str__ = __repr__



from .audit_log import *
from .audit_log_change import *
from .audit_log_entry import *
from .audit_log_entry_change_conversion import *
from .audit_log_entry_change_conversions import *
from .audit_log_entry_detail_conversion import *
from .audit_log_entry_detail_conversions import *
from .audit_log_iterator import *
from .audit_log_role import *
from .conversion_helpers import *


__all__ = (
    *audit_log.__all__,
    *audit_log_change.__all__,
    *audit_log_entry.__all__,
    *audit_log_entry_change_conversion.__all__,
    *audit_log_entry_change_conversions.__all__,
    *audit_log_entry_detail_conversion.__all__,
    *audit_log_entry_detail_conversions.__all__,
    *audit_log_iterator.__all__,
    *audit_log_role.__all__,
    *conversion_helpers.__all__,
)



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_optional_putter_factory
from ...field_validators import nullable_string_validator_factory

from ..component import Component, ComponentType, ComponentTypeLayoutFlag
from ..shared_fields import parse_components, parse_custom_id, put_components, put_custom_id, validate_custom_id

from .constants import TITLE_LENGTH_MAX, TITLE_LENGTH_MIN


COMPONENT_TYPE_LAYOUT_FLAG_TOP_LEVEL = ComponentTypeLayoutFlag().update_by_keys(top_level = True)
COMPONENT_TYPE_LAYOUT_FLAG_ALLOWED_IN_FORM = ComponentTypeLayoutFlag().update_by_keys(allowed_in_form = True)


# components

def validate_components(components):
    """
    Interaction form specific component validator.
    
    Parameters
    ----------
    components : ``None | iterable<Component | (tuple | list)<Component>>``
        The components to validate.
    
    Returns
    -------
    components : ``None | tuple<Component>``
    
    Raises
    ------
    TypeError
        - If `components` is invalid type.
    ValueError
        - If `components` contains an invalid value.
    """
    if components is None:
        return None
    
    if (getattr(components, '__iter__', None) is None):
        raise TypeError(
            f'`components` can be `None`, `iterable` of `{Component.__name__}`, got '
            f'{type(components).__name__}; {components!r}.'
        )
        
    components_processed = None
    
    for component in components:
        if not isinstance(component, Component):
            raise TypeError(
                f'`components` can contain `{Component.__name__}` elements, got '
                f'{type(component).__name__}; {component!r}; components = {components!r}.'
            )
        
        flags = component.type.layout_flags
        if not flags & COMPONENT_TYPE_LAYOUT_FLAG_ALLOWED_IN_FORM:
            raise ValueError(
                f'Component {component!r} is not allowed to be in a form.',
            )
        
        if not flags & COMPONENT_TYPE_LAYOUT_FLAG_TOP_LEVEL:
            label = component.label
            if (label is not None):
                component.label = None
            
            component = Component(
                ComponentType.label,
                component = component,
                label = label,
            )
        
        if (components_processed is None):
            components_processed = []
        
        components_processed.append(component)
    
    if (components_processed is not None):
        components_processed = tuple(components_processed)
    
    return components_processed


# title

parse_title = nullable_string_parser_factory('title')
put_title = nullable_string_optional_putter_factory('title')
validate_title = nullable_string_validator_factory('title', TITLE_LENGTH_MIN, TITLE_LENGTH_MAX)



__all__ = ('InteractionMetadataFormSubmit',)

from scarletio import copy_docs

from .base import InteractionMetadataBase
from .fields import (
    parse_custom_id, parse_components, put_custom_id, put_components, validate_custom_id, validate_components
)


class InteractionMetadataFormSubmit(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Parameters
    ----------
    components : ``None | tuple<InteractionComponent>``
        Submitted component values of a form submit interaction.
    
    custom_id : `None | str`
        Form interaction's custom identifier.
    """
    __slots__ = ('components', 'custom_id')
    
    def __new__(
        cls,
        *,
        components = ...,
        custom_id = ...,
    ):
        """
        Creates a new interaction metadata from the given parameters.
        
        Parameters
        ----------
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None | str`, Optional (Keyword only)
            Form interaction's custom identifier.

        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # components
        if components is ...:
            components = None
        else:
            components = validate_components(components)
        
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # Construct
        self = object.__new__(cls)
        self.components = components
        self.custom_id = custom_id
        return self
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            components = keyword_parameters.pop('components', ...),
            custom_id = keyword_parameters.pop('custom_id', ...),
        )
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.components = None
        self.custom_id = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        components = self.components
        if (components is not None):
            components = (*(option.copy() for option in components),)
        new.components = components
        
        new.custom_id = self.custom_id
        
        return new
    
    
    def copy_with(
        self,
        *,
        components = ...,
        custom_id = ...,
    ):
        """
        Copies the interaction metadata with the given fields.
        
        Parameters
        ----------
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None | str`, Optional (Keyword only)
            Form interaction's custom identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # components
        if components is ...:
            components = self.components
            if (components is not None):
                components = (*(option.copy() for option in components),)
        else:
            components = validate_components(components)
        
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # Construct
        new = object.__new__(type(self))
        new.components = components
        new.custom_id = custom_id
        return new
    
    
    @copy_docs(InteractionMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            components = keyword_parameters.pop('components', ...),
            custom_id = keyword_parameters.pop('custom_id', ...),
        )
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, guild_id = 0):
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.custom_id = parse_custom_id(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, guild_id = 0):
        data = {}
        put_components(self.components, data, defaults)
        put_custom_id(self.custom_id, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        
        # components
        components = self.components
        if (components is not None):
            repr_parts.append(' components = ')
            repr_parts.append(repr(components))
            
            field_added = False
        else:
            field_added = True
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' custom_id = ')
            repr_parts.append(repr(custom_id))
        
        return field_added
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components)
            
            for option in components:
                hash_value ^= hash(option)
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        return True
    
    
    @copy_docs(InteractionMetadataBase.iter_components)
    def iter_components(self):
        components = self.components
        if (components is not None):
            yield from components
    
    
    @copy_docs(InteractionMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        components = self.components
        if (components is not None):
            for component in components:
                yield from component.iter_custom_ids_and_values()



__all__ = (
    'Typer', 'run_console_till_interruption', 'start_clients', 'stop_clients', 'wait_for_interruption'
)

import sys
from threading import current_thread
from time import sleep as blocking_sleep

from scarletio import CancelledError, Task, TaskGroup, get_last_module_frame, sleep
from scarletio.tools.asynchronous_interactive_console import (
    create_banner, create_exit_message, run_asynchronous_interactive_console
)

from ... import __package__ as PACKAGE_NAME
from ..core import CLIENTS, KOKORO
from ..permission import Permission
from ..permission.constants import PERMISSION_KEY

PACKAGE = __import__(PACKAGE_NAME)


def start_clients():
    """
    Starts up all the not running clients.
    
    Can be called from any thread.
    """
    for client in CLIENTS.values():
        if client.running:
            continue
        
        Task(KOKORO, client.connect())
    
    if (current_thread() is not KOKORO):
        KOKORO.wake_up()


def stop_clients():
    """
    Stops all the running clients.
    
    Can be called from any thread.
    """
    for client in CLIENTS.values():
        if client.running:
            Task(KOKORO, client.disconnect())
    
    if (current_thread() is not KOKORO):
        KOKORO.wake_up()


def wait_for_interruption():
    """
    Waits for keyboard interruption. If occurred, shuts down all the clients and closes the event loop. This operation
    is unreleasable.
    
    > Shutting down the clients might take a few seconds depending on the added shutdown event handlers.
    
    Raises
    ------
    RuntimeError
        If used inside of the event loop of clients.
    KeyboardInterrupt
        The received keyboard interrupt.
    """
    if current_thread() is KOKORO:
        raise RuntimeError(f'`wait_for_interruption` cannot be used inside of {KOKORO!r}.')
    
    try:
        while True:
            # sleep 1 day
            blocking_sleep(86400.0)
    except KeyboardInterrupt as err:
        exception = err
    else:
        # should not happen
        exception = None
    
    _exit_callback()
    
    KOKORO.stop()
    
    # reraise exception
    if (exception is not None):
        raise exception


def _exit_callback():
    """
    Callback used by ``wait_for_interruption`` and ``run_console_till_interruption``
    to stop the running clients.
    """
    sys.stdout.write('\ninterrupted ...\n')
    
    TaskGroup(
        KOKORO,
        (Task(KOKORO, client.disconnect()) for client in CLIENTS.values())
    ).wait_all().sync_wrap().wait()


IGNORED_CONSOLE_VARIABLES = {'__name__', '__package__', '__loader__', '__spec__'}

def run_console_till_interruption():
    """
    Runs interactive console.
    
    On system exit shuts down all the clients and closes the event loop.
    
    > Shutting down the clients might take a few seconds depending on the added shutdown event handlers.
    
    Raises
    ------
    RuntimeError
        If used inside of the event loop of clients.
    """
    if current_thread() is KOKORO:
        raise RuntimeError(f'`run_console_till_interruption` cannot be used inside of {KOKORO!r}.')
    
    frame = get_last_module_frame()
    
    interactive_console_locals = {}
    for variable_name, variable_value in frame.f_globals.items():
        if (variable_name not in IGNORED_CONSOLE_VARIABLES):
            interactive_console_locals[variable_name] = variable_value
    
    run_asynchronous_interactive_console(
        interactive_console_locals,
        banner = create_banner(PACKAGE),
        exit_message = create_exit_message(PACKAGE),
        callback = _exit_callback,
        stop_on_interruption = True,
    )


class UserGuildPermission:
    """
    Represents a user's permissions inside of a guild. Returned by ``Client.user_guild_get_all``.
    
    Attributes
    ----------
    owner : `bool`
        Whether the user is the owner of the guild.
    permission : ``Permission``
        The user's permissions at the guild.
    """
    __slots__ = ('owner', 'permission', )
    
    def __init__(self, data):
        """
        Creates a ``GuildPermission`` object form user guild data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild permission data.
        """
        self.owner = data['owner']
        self.permission = Permission(data[PERMISSION_KEY])
    
    
    def __repr__(self):
        """Returns the user guild permission's representation."""
        return f'<{self.__class__.__name__}  owner = {self.owner}, permissions = {self.permission:d}>'
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 2
    
    
    def __iter__(self):
        """Unpacks the user guild permission."""
        yield self.owner
        yield self.permission


class Typer:
    """
    A typer what will keep sending typing events to the given channel with the client. Can be used as a context
    manager.
    
    After entered as a context manager sends a typing event each `8` seconds to the given channel.
    
    Attributes
    ----------
    client : ``Client``
        The client what will send the typing events.
    channel_id : `int`
        The channel's id where typing will be triggered.
    timeout : `float`
        The leftover timeout till the typer will send typing events. Is reduced every time, when the typer sent a typing
        event. If goes under `0.0` the typer stops sending more events.
    waiter : ``Future``, `None`
        The sleeping future what will wake_up ``.run``.
    """
    __slots__ = ('channel_id', 'client', 'timeout', 'waiter',)
    
    def __init__(self, client, channel_id, timeout = 300.0):
        """
        Parameters
        ----------
        client : ``Client``
            The client what will send the typing events.
        channel_id : `int`
            The channel's id where typing will be triggered.
        timeout : `float` = `300.0`, Optional
            The maximal amount of time till the client will keep sending typing events. Defaults to `300.0`.
        """
        self.client = client
        self.channel_id = channel_id
        self.waiter = None
        self.timeout = timeout
    
    
    def __enter__(self):
        """Enters the typer's context block by ensuring it's ``.run`` method."""
        Task(KOKORO, self.run())
        return self
    
    
    async def run(self):
        """
        The coroutine what keeps sending the typing requests.
        
        This method is a coroutine.
        """
        # js client's typing is 8s
        while self.timeout > 0.0:
            self.timeout -= 8.0
            self.waiter = waiter = sleep(8.0, KOKORO)
            await self.client.api.typing(self.channel_id)
            await waiter
        
        self.waiter = None
    
    
    def __await__(self):
        """Keeps typing till timeout occurs."""
        while True:
            timeout = self.timeout
            if timeout <= 0.0:
                break
            
            self.timeout = new_timeout = timeout - 8.0
            if new_timeout < 0.0:
                sleep_duration = 8.0 + new_timeout
            else:
                sleep_duration = 8.0
            
            self.waiter = waiter = sleep(sleep_duration, KOKORO)
            yield from self.client.api.typing(self.channel_id).__await__()
            
            try:
                yield from waiter
            except CancelledError:
                # Reraise if cancelled from outside
                if (self.waiter is not None):
                    raise
    
    
    def cancel(self):
        """
        If the context manager is still active, cancels it.
        """
        self.timeout = 0.0
        waiter = self.waiter
        if (waiter is not None):
            self.waiter = None
            waiter.cancel()
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """Exits the typer's context block by cancelling it."""
        self.cancel()
        return False
